# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import sys
from typing import TypeVar, Type, Sequence, Union, cast
import struct
import base64

import numpy
from numpy.typing import NDArray
from ._common import StdPrimitive, Byte

# noinspection PyShadowingBuiltins
_T = TypeVar("_T")


class Deserializer(abc.ABC):
    """
    The deserializer class is used for deconstruction of serialized representations of DSDL objects into Python objects.
    It implements the implicit zero extension rule as described in the Specification.
    """

    class FormatError(ValueError):
        """
        This exception class is used when an auto-generated deserialization routine is supplied with invalid input data;
        in other words, input that is not a valid serialized representation of its data type.

        Deserialization logic (auto-generated or manually written) may use this exception type.
        When thrown from a deserialization method, it is intercepted by :func:`pyuavcan.dsdl.deserialize`
        which then returns None instead of a valid instance, indicating that the serialized representation is invalid.
        """

    def __init__(self, fragmented_buffer: Sequence[memoryview]):
        """
        Do not call this directly. Use :meth:`new` to instantiate.
        """
        self._buf = ZeroExtendingBuffer(fragmented_buffer)
        self._bit_offset = 0
        assert self.consumed_bit_length + self.remaining_bit_length == self._buf.bit_length

    @staticmethod
    def new(fragmented_buffer: Sequence[memoryview]) -> Deserializer:
        """
        :param fragmented_buffer: The source serialized representation. The deserializer will attempt to avoid copying
            any data from the serialized representation, establishing direct references to its memory instead.
            If any of the source buffer fragments are read-only, some of the deserialized array-typed values
            may end up being read-only as well. If that is undesirable, use writeable buffer.

        :return: A new instance of Deserializer, either little-endian or big-endian, depending on the platform.
        """
        return _PlatformSpecificDeserializer(fragmented_buffer)

    @property
    def consumed_bit_length(self) -> int:
        return self._bit_offset

    @property
    def remaining_bit_length(self) -> int:
        """Returns negative if out of bounds (zero extension rule in effect)."""
        return self._buf.bit_length - self._bit_offset

    def skip_bits(self, bit_length: int) -> None:
        """This is used for padding bits."""
        _ensure_cardinal(bit_length)
        self._bit_offset += bit_length

    def pad_to_alignment(self, bit_length: int) -> None:
        while self._bit_offset % bit_length != 0:
            self._bit_offset += 1

    def fork_bytes(self, forked_buffer_size_in_bytes: int) -> Deserializer:
        """
        This is the counterpart of fork_bytes() defined in the serializer intended for deserializing delimited types.
        Forking is necessary to support implicit truncation and implicit zero extension of nested objects.
        The algorithm is as follows:

        - Before forking, using the main deserializer (M), read the delimiter header.
        - If the value of the delimiter header exceeds the number of bytes remaining in the deserialization buffer,
          raise :class:`FormatError`, thereby declaring the serialized representation invalid, as prescribed by the
          Specification.
        - Fork M.
        - Skip M by the size reported by the delimiter header.
        - Using the forked deserializer (F), deserialize the nested object. F will apply implicit truncation
          and the implicit zero extension rules as necessary regardless of the amount of data remaining in M.
        - Discard F.

        This method raises a :class:`ValueError` if the forked instance is not byte-aligned or if the requested buffer
        size is too large. The latter is because it is a class usage error, not a deserialization error.
        """
        if self._bit_offset % 8 != 0:
            raise ValueError("Cannot fork unaligned deserializer")
        remaining_bit_length = self.remaining_bit_length
        assert remaining_bit_length % 8 == 0
        remaining_byte_length = remaining_bit_length // 8
        if remaining_byte_length < forked_buffer_size_in_bytes:
            raise ValueError(
                f"Invalid usage: the required forked buffer size of {forked_buffer_size_in_bytes} bytes "
                f"is less than the available remaining buffer space of {remaining_byte_length} bytes"
            )
        out = _PlatformSpecificDeserializer(self._buf.fork_bytes(self._byte_offset, forked_buffer_size_in_bytes))
        assert out.remaining_bit_length == forked_buffer_size_in_bytes * 8
        return out

    #
    # Fast methods optimized for aligned primitive fields.
    # The most specialized methods must be used whenever possible for best performance.
    #
    @abc.abstractmethod
    def fetch_aligned_array_of_standard_bit_length_primitives(
        self, dtype: Type[StdPrimitive], count: int
    ) -> NDArray[StdPrimitive]:
        """
        Returns a new array which may directly refer to the underlying memory.
        The returned array may be read-only if the source buffer is read-only.
        """
        raise NotImplementedError

    def fetch_aligned_array_of_bits(self, count: int) -> NDArray[numpy.bool_]:
        """
        Quickly decodes an aligned array of bits using the numpy's fast bit unpacking routine.
        A new array is always created (the memory cannot be shared with the buffer due to the layout transformation).
        The returned array is of dtype :class:`bool`.
        """
        _ensure_cardinal(count)
        assert self._bit_offset % 8 == 0
        bs = self._buf.get_unsigned_slice(self._byte_offset, self._byte_offset + (count + 7) // 8)
        out = numpy.unpackbits(bs, bitorder="little")[:count]
        self._bit_offset += count
        assert len(out) == count
        return cast(NDArray[numpy.bool_], out.astype(dtype=bool))

    def fetch_aligned_bytes(self, count: int) -> NDArray[Byte]:
        _ensure_cardinal(count)
        assert self._bit_offset % 8 == 0
        out = self._buf.get_unsigned_slice(self._byte_offset, self._byte_offset + count)
        self._bit_offset += count * 8
        assert len(out) == count
        return out

    def fetch_aligned_u8(self) -> int:
        assert self._bit_offset % 8 == 0
        out = self._buf.get_byte(self._byte_offset)
        assert isinstance(out, int)  # Make sure it's not a NumPy's integer type like numpy.uint8. We need native int.
        self._bit_offset += 8
        return out

    def fetch_aligned_u16(self) -> int:  # TODO: here and below, consider using int.from_bytes()?
        out = self.fetch_aligned_u8()
        out |= self.fetch_aligned_u8() << 8
        return out

    def fetch_aligned_u32(self) -> int:
        out = self.fetch_aligned_u16()
        out |= self.fetch_aligned_u16() << 16
        return out

    def fetch_aligned_u64(self) -> int:
        out = self.fetch_aligned_u32()
        out |= self.fetch_aligned_u32() << 32
        return out

    def fetch_aligned_i8(self) -> int:
        x = self.fetch_aligned_u8()
        return (x - 256) if x >= 128 else x

    def fetch_aligned_i16(self) -> int:
        x = self.fetch_aligned_u16()
        return (x - 65536) if x >= 32768 else x

    def fetch_aligned_i32(self) -> int:
        x = self.fetch_aligned_u32()
        return int(x - 2 ** 32) if x >= 2 ** 31 else x  # wrapped in int() to appease MyPy

    def fetch_aligned_i64(self) -> int:
        x = self.fetch_aligned_u64()
        return int(x - 2 ** 64) if x >= 2 ** 63 else x  # wrapped in int() to appease MyPy

    def fetch_aligned_f16(self) -> float:  # noinspection PyTypeChecker
        (out,) = struct.unpack("<e", self.fetch_aligned_bytes(2))  # type: ignore
        assert isinstance(out, float)
        return out

    def fetch_aligned_f32(self) -> float:  # noinspection PyTypeChecker
        (out,) = struct.unpack("<f", self.fetch_aligned_bytes(4))  # type: ignore
        assert isinstance(out, float)
        return out

    def fetch_aligned_f64(self) -> float:  # noinspection PyTypeChecker
        (out,) = struct.unpack("<d", self.fetch_aligned_bytes(8))  # type: ignore
        assert isinstance(out, float)
        return out

    #
    # Less specialized methods: assuming that the value is aligned at the beginning, but its bit length
    # is non-standard and may not be an integer multiple of eight.
    # These must not be used if there is a suitable more specialized version defined above.
    #
    def fetch_aligned_unsigned(self, bit_length: int) -> int:
        _ensure_cardinal(bit_length)
        assert self._bit_offset % 8 == 0
        bs = self._buf.get_unsigned_slice(self._byte_offset, self._byte_offset + (bit_length + 7) // 8)
        self._bit_offset += bit_length
        return self._unsigned_from_bytes(bs, bit_length)

    def fetch_aligned_signed(self, bit_length: int) -> int:
        assert bit_length >= 2
        u = self.fetch_aligned_unsigned(bit_length)
        out = (u - 2 ** bit_length) if u >= 2 ** (bit_length - 1) else u
        assert isinstance(out, int)  # MyPy pls
        return out

    #
    # Least specialized methods: no assumptions about alignment are made.
    # These are the slowest and may be used only if none of the above (specialized) methods are suitable.
    #
    @abc.abstractmethod
    def fetch_unaligned_array_of_standard_bit_length_primitives(
        self, dtype: Type[StdPrimitive], count: int
    ) -> NDArray[StdPrimitive]:
        """See the aligned counterpart."""
        raise NotImplementedError

    def fetch_unaligned_array_of_bits(self, count: int) -> NDArray[numpy.bool_]:
        _ensure_cardinal(count)
        byte_count = (count + 7) // 8
        bs = self.fetch_unaligned_bytes(byte_count)
        assert len(bs) == byte_count
        backtrack = byte_count * 8 - count
        assert 0 <= backtrack < 8
        self._bit_offset -= backtrack
        out: NDArray[numpy.bool_] = numpy.unpackbits(bs, bitorder="little")[:count].astype(dtype=bool)
        assert len(out) == count
        return out

    def fetch_unaligned_bytes(self, count: int) -> NDArray[Byte]:
        if count > 0:
            if self._bit_offset % 8 != 0:
                # This is a faster variant of Ben Dyer's unaligned bit copy algorithm:
                # https://github.com/UAVCAN/libuavcan/blob/fd8ba19bc9c09/libuavcan/src/marshal/uc_bit_array_copy.cpp#L12
                # It is faster because here we are aware that the destination is always aligned, which we take
                # advantage of. This algorithm breaks for byte-aligned offset, so we have to delegate the aligned
                # case to the aligned copy method (which is also much faster).
                out: NDArray[Byte] = numpy.empty(count, dtype=Byte)
                right = self._bit_offset % 8
                left = 8 - right
                assert (1 <= right <= 7) and (1 <= left <= 7)
                # The last byte is a special case because if we're reading the last few unaligned bits, the very last
                # byte access will be always out of range. We don't care because of the implicit zero extension rule.
                for i in range(count):
                    byte_offset = self._byte_offset
                    out[i] = (self._buf.get_byte(byte_offset) >> right) | (
                        (self._buf.get_byte(byte_offset + 1) << left) & 0xFF
                    )
                    self._bit_offset += 8
                assert len(out) == count
                return out
            return self.fetch_aligned_bytes(count)
        return numpy.zeros(0, dtype=Byte)

    def fetch_unaligned_unsigned(self, bit_length: int) -> int:
        _ensure_cardinal(bit_length)
        byte_length = (bit_length + 7) // 8
        bs = self.fetch_unaligned_bytes(byte_length)
        assert len(bs) == byte_length
        backtrack = byte_length * 8 - bit_length
        assert 0 <= backtrack < 8
        self._bit_offset -= backtrack
        return self._unsigned_from_bytes(bs, bit_length)

    def fetch_unaligned_signed(self, bit_length: int) -> int:
        assert bit_length >= 2
        u = self.fetch_unaligned_unsigned(bit_length)
        out = (u - 2 ** bit_length) if u >= 2 ** (bit_length - 1) else u
        assert isinstance(out, int)  # MyPy pls
        return out

    def fetch_unaligned_f16(self) -> float:  # noinspection PyTypeChecker
        (out,) = struct.unpack("<e", self.fetch_unaligned_bytes(2))  # type: ignore
        assert isinstance(out, float)
        return out

    def fetch_unaligned_f32(self) -> float:  # noinspection PyTypeChecker
        (out,) = struct.unpack("<f", self.fetch_unaligned_bytes(4))  # type: ignore
        assert isinstance(out, float)
        return out

    def fetch_unaligned_f64(self) -> float:  # noinspection PyTypeChecker
        (out,) = struct.unpack("<d", self.fetch_unaligned_bytes(8))  # type: ignore
        assert isinstance(out, float)
        return out

    def fetch_unaligned_bit(self) -> bool:
        mask = 1 << (self._bit_offset % 8)
        assert 1 <= mask <= 128
        out = self._buf.get_byte(self._byte_offset) & mask == mask
        self._bit_offset += 1
        return bool(out)

    #
    # Private methods.
    #
    @staticmethod
    def _unsigned_from_bytes(x: NDArray[Byte], bit_length: int) -> int:
        assert bit_length >= 1
        num_bytes = (bit_length + 7) // 8
        assert num_bytes > 0
        last_byte_index = num_bytes - 1
        assert len(x) >= num_bytes
        out = 0
        for i in range(last_byte_index):
            out |= int(x[i]) << (i * 8)
        msb_mask = (2 ** (bit_length % 8) - 1) if bit_length % 8 != 0 else 0xFF
        assert msb_mask in (1, 3, 7, 15, 31, 63, 127, 255)
        out |= (int(x[last_byte_index]) & msb_mask) << (last_byte_index * 8)
        assert 0 <= out < (2 ** bit_length)
        return out

    @property
    def _byte_offset(self) -> int:
        return self._bit_offset // 8

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"consumed_bit_length={self.consumed_bit_length}, "
            f"remaining_bit_length={self.remaining_bit_length}, "
            f"serialized_representation_base64={self._buf.to_base64()!r})"
        )


class _LittleEndianDeserializer(Deserializer):
    def fetch_aligned_array_of_standard_bit_length_primitives(
        self, dtype: Type[StdPrimitive], count: int
    ) -> NDArray[StdPrimitive]:
        assert dtype not in (bool, numpy.bool_, object), "Invalid usage"
        assert self._bit_offset % 8 == 0
        bo = self._byte_offset
        # Interestingly, numpy doesn't care about alignment. If the source buffer is not properly aligned, it will
        # work anyway but slower.
        out: NDArray[StdPrimitive] = numpy.frombuffer(
            self._buf.get_unsigned_slice(bo, bo + count * numpy.dtype(dtype).itemsize), dtype=dtype
        )
        assert len(out) == count
        self._bit_offset += out.nbytes * 8
        return out

    def fetch_unaligned_array_of_standard_bit_length_primitives(
        self, dtype: Type[StdPrimitive], count: int
    ) -> NDArray[StdPrimitive]:
        assert dtype not in (bool, numpy.bool_, object), "Invalid usage"
        bs = self.fetch_unaligned_bytes(numpy.dtype(dtype).itemsize * count)
        assert len(bs) >= count
        return numpy.frombuffer(bs, dtype=dtype, count=count)


class _BigEndianDeserializer(Deserializer):
    def fetch_aligned_array_of_standard_bit_length_primitives(
        self, dtype: Type[StdPrimitive], count: int
    ) -> NDArray[StdPrimitive]:
        raise NotImplementedError("Pull requests are welcome")

    def fetch_unaligned_array_of_standard_bit_length_primitives(
        self, dtype: Type[StdPrimitive], count: int
    ) -> NDArray[StdPrimitive]:
        raise NotImplementedError("Pull requests are welcome")


_PlatformSpecificDeserializer = {
    "little": _LittleEndianDeserializer,
    "big": _BigEndianDeserializer,
}[sys.byteorder]


class ZeroExtendingBuffer:
    """
    This class implements the implicit zero extension logic as described in the Specification.
    A read beyond the end of the buffer returns zero bytes.
    """

    def __init__(self, fragmented_buffer: Sequence[memoryview]):
        # TODO: Concatenation is a tentative measure. Add proper support for fragmented buffers for speed.
        if len(fragmented_buffer) == 1:
            contiguous: Union[bytearray, memoryview] = fragmented_buffer[0]  # Fast path.
        else:
            contiguous = bytearray().join(fragmented_buffer)

        self._buf: NDArray[Byte] = numpy.frombuffer(contiguous, dtype=Byte)
        assert self._buf.dtype == Byte and self._buf.ndim == 1

    @property
    def bit_length(self) -> int:
        return len(self._buf) * 8

    def get_byte(self, index: int) -> int:
        """
        Like the standard ``x[i]`` except that i may not be negative and out of range access returns zero.
        """
        if index < 0:
            raise ValueError("Byte index may not be negative because the end of a zero-extended buffer is undefined.")
        try:
            return int(self._buf[index])
        except IndexError:
            return 0  # Implicit zero extension rule

    def get_unsigned_slice(self, left: int, right: int) -> NDArray[Byte]:
        """
        Like the standard ``x[left:right]`` except that neither index may be negative,
        left may not exceed right (otherwise it's a :class:`ValueError`),
        and the returned value is always of size ``right-left`` right-zero-padded if necessary.
        """
        if not (0 <= left <= right):
            raise ValueError(f"Invalid slice boundary specification: [{left}:{right}]")
        count = int(right - left)
        assert count >= 0
        out: NDArray[Byte] = self._buf[left:right]  # Slicing never raises an IndexError.
        if len(out) < count:  # Implicit zero extension rule
            out = numpy.concatenate((out, numpy.zeros(count - len(out), dtype=Byte)))
        assert len(out) == count
        return out

    def fork_bytes(self, offset_bytes: int, length_bytes: int) -> Sequence[memoryview]:
        """
        This is intended for use with :meth:`Deserializer.fork_bytes`.
        Given an offset from the beginning and length (both in bytes), yields a list of compliant memory fragments
        that can be fed into the forked deserializer instance.
        The requested (offset + length) shall not exceeded the buffer length; this is because per the Specification,
        a delimiter header cannot exceed the amount of remaining space in the deserialization buffer.
        """
        # Currently, we use a contiguous buffer, but when scattered buffers are supported, this method will need
        # to discard the fragments before the requested offset and then return the following subset of fragments.
        if offset_bytes + length_bytes > len(self._buf):
            raise ValueError(f"Invalid fork: offset ({offset_bytes}) + length ({length_bytes}) > {len(self._buf)}")
        out = memoryview(self._buf[offset_bytes : offset_bytes + length_bytes])
        assert len(out) == length_bytes
        return [out]

    def to_base64(self) -> str:
        return base64.b64encode(self._buf.tobytes()).decode()


def _ensure_cardinal(i: int) -> None:
    if i < 0:
        raise ValueError(f"Cardinal may not be negative: {i}")


def _unittest_deserializer_aligned() -> None:
    from pytest import raises, approx

    # The buffer is constructed from the corresponding serialization test.
    # The final bit padding is done with 1's to ensure that they are correctly discarded.
    sample = bytes(
        map(
            lambda x: int(x, 2),
            "10100111 11101111 11001101 10101011 10010000 01111000 01010110 00110100 00010010 10001000 10101001 "
            "11001011 11101101 11111110 11111111 00000000 01111111 00000000 00000000 00000000 00000000 00000000 "
            "00000000 11110000 00111111 00000000 00000000 10000000 00111111 00000000 01111100 11011010 00001110 "
            "11011010 10111110 11111110 00000001 10101101 11011110 11101111 10111110 11000101 01100111 11000101 "
            "11101011".split(),
        )
    )
    assert len(sample) == 45

    des = Deserializer.new([memoryview(sample)])
    assert des.remaining_bit_length == 45 * 8

    assert des.fetch_aligned_u8() == 0b1010_0111
    assert des.fetch_aligned_i64() == 0x1234_5678_90AB_CDEF
    assert des.fetch_aligned_i32() == -0x1234_5678
    assert des.fetch_aligned_i16() == -2

    assert des.remaining_bit_length == 45 * 8 - 8 - 64 - 32 - 16
    des.skip_bits(8)
    assert des.remaining_bit_length == 45 * 8 - 8 - 64 - 32 - 16 - 8

    assert des.fetch_aligned_i8() == 127
    assert des.fetch_aligned_f64() == approx(1.0)
    assert des.fetch_aligned_f32() == approx(1.0)
    assert des.fetch_aligned_f16() == numpy.inf

    assert des.fetch_aligned_unsigned(12) == 0xEDA
    des.skip_bits(4)
    assert des.fetch_aligned_unsigned(16) == 0xBEDA
    assert des.fetch_aligned_signed(9) == -2
    des.skip_bits(7)

    assert all(des.fetch_aligned_array_of_standard_bit_length_primitives(numpy.uint16, 2) == [0xDEAD, 0xBEEF])

    assert all(
        des.fetch_aligned_array_of_bits(16)
        == [
            True,
            False,
            True,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
            True,
            True,
            False,
        ]
    )

    assert all(
        des.fetch_aligned_array_of_bits(13)
        == [
            True,
            False,
            True,
            False,
            False,
            False,
            True,
            True,
            True,
            True,
            False,
            True,
            False,
        ]
    )

    print("repr(deserializer):", repr(des))

    des = Deserializer.new([memoryview(bytes([1, 2, 3]))])

    assert list(des.fetch_aligned_array_of_bits(0)) == []
    assert list(des.fetch_aligned_bytes(0)) == []
    assert des.remaining_bit_length == 3 * 8

    with raises(ValueError):
        des.fetch_aligned_array_of_bits(-1)

    with raises(ValueError):
        des.fetch_aligned_bytes(-1)

    des.skip_bits(3 * 8)
    assert des.remaining_bit_length == 0

    assert all([False] * 100 == des.fetch_aligned_array_of_bits(100))  # type: ignore
    assert des.remaining_bit_length == -100
    des.skip_bits(4)
    assert des.remaining_bit_length == -104
    assert b"\x00" * 10 == des.fetch_aligned_bytes(10).tobytes()
    assert des.remaining_bit_length == -184
    des.skip_bits(64)
    assert des.remaining_bit_length == -248
    assert 0 == des.fetch_aligned_unsigned(64)
    assert des.remaining_bit_length == -312

    print("repr(deserializer):", repr(des))


def _unittest_deserializer_unaligned() -> None:
    from pytest import approx

    des = Deserializer.new([memoryview(bytearray([0b10101010, 0b01011101, 0b11001100, 0b10010001]))])
    assert des.consumed_bit_length == 0
    assert des.consumed_bit_length % 8 == 0
    assert list(des.fetch_aligned_array_of_bits(3)) == [False, True, False]
    assert des.consumed_bit_length == 3
    assert des.consumed_bit_length % 8 == 3
    assert list(des.fetch_unaligned_bytes(0)) == []
    assert list(des.fetch_unaligned_bytes(2)) == [0b10110101, 0b10001011]
    assert list(des.fetch_unaligned_bytes(1)) == [0b00111001]
    assert des.consumed_bit_length == 27
    assert des.consumed_bit_length % 8 == 3
    assert des.remaining_bit_length == 5
    assert all(numpy.array([0b00010010, 0], dtype=Byte) == des.fetch_unaligned_bytes(2))
    assert des.consumed_bit_length == 43
    assert des.remaining_bit_length == -11

    des = Deserializer.new([memoryview(bytearray([0b10101010, 0b01011101, 0b11001100, 0b10010001]))])
    assert list(des.fetch_unaligned_bytes(0)) == []
    assert list(des.fetch_unaligned_bytes(2)) == [0b10101010, 0b01011101]  # Actually aligned
    assert list(des.fetch_unaligned_bytes(1)) == [0b11001100]
    assert des.remaining_bit_length == 8
    assert list(des.fetch_unaligned_bytes(2)) == [0b10010001, 0]
    assert des.remaining_bit_length == -8

    # The buffer is constructed from the corresponding serialization test.
    sample = bytearray(
        map(
            lambda x: int(x, 2),
            "11000101 00101111 01010111 10000010 11000110 11001010 00010010 00110100 01010110 11011001 10111111 "
            "11101100 00000110 00000000 00000000 00000000 00000000 00000000 10000000 11111111 00000001 00000000 "
            "00000000 11111100 00000001 11100000 01101111 11110101 01111110 11110111 00000101".split(),
        )
    )
    assert len(sample) == 31

    des = Deserializer.new([memoryview(sample[:])])
    assert des.remaining_bit_length == 31 * 8

    assert list(des.fetch_unaligned_array_of_bits(11)) == [
        True,
        False,
        True,
        False,
        False,
        False,
        True,
        True,  # 10100011
        True,
        True,
        True,  # 111
    ]
    assert list(des.fetch_unaligned_array_of_bits(10)) == [
        True,
        False,
        True,
        False,
        False,  # ???10100 (byte alignment restored here)
        True,
        True,
        True,
        False,
        True,  # 11101 (byte alignment lost, three bits short)
    ]

    assert list(des.fetch_unaligned_bytes(3)) == [0x12, 0x34, 0x56]
    assert list(des.fetch_unaligned_array_of_bits(3)) == [False, True, True]
    assert list(des.fetch_unaligned_bytes(3)) == [0x12, 0x34, 0x56]

    assert des.fetch_unaligned_bit()
    assert not des.fetch_unaligned_bit()
    assert not des.fetch_unaligned_bit()
    assert des.fetch_unaligned_bit()
    assert des.fetch_unaligned_bit()

    assert des.fetch_unaligned_signed(8) == -2
    assert des.fetch_unaligned_unsigned(11) == 0b111_0110_0101
    assert des.fetch_unaligned_unsigned(3) == 0b110

    assert des.consumed_bit_length % 8 > 0  # not aligned
    assert des.fetch_unaligned_f64() == approx(1.0)
    assert des.fetch_unaligned_f32() == approx(1.0)
    assert des.fetch_unaligned_f16() == -numpy.inf

    assert list(des.fetch_unaligned_array_of_standard_bit_length_primitives(numpy.uint16, 2)) == [0xDEAD, 0xBEEF]
    des.skip_bits(5)
    assert des.consumed_bit_length % 8 == 0
    assert des.remaining_bit_length == 0

    print("repr(deserializer):", repr(des))


def _unittest_deserializer_fork_bytes() -> None:
    import pytest

    m = Deserializer.new(
        [
            memoryview(
                bytes([0b10100111, 0b11101111, 0b11001101, 0b10101011, 0b10010000, 0b01111000, 0b01010110, 0b00110100])
            )
        ]
    )
    with pytest.raises(ValueError):
        m.fork_bytes(9)

    f = m.fork_bytes(8)
    assert f.consumed_bit_length == 0
    assert f.remaining_bit_length == 8 * 8
    assert f.fetch_aligned_u8() == 0b10100111
    assert f.remaining_bit_length == 7 * 8
    assert f.fetch_aligned_u8() == 0b11101111
    assert f.remaining_bit_length == 6 * 8
    assert f.consumed_bit_length == 16

    assert m.remaining_bit_length == 8 * 8
    m.skip_bits(6 * 8)
    assert m.remaining_bit_length == 2 * 8
    assert m.fetch_aligned_u8() == 0b01010110
    assert m.fetch_aligned_u8() == 0b00110100
    assert m.remaining_bit_length == 0
    assert m.fetch_aligned_u8() == 0
    assert m.fetch_aligned_u16() == 0
    assert m.fetch_aligned_u32() == 0
    assert m.fetch_aligned_u64() == 0

    assert f.remaining_bit_length == 6 * 8
    ff = f.fork_bytes(2)
    assert ff.consumed_bit_length == 0
    assert ff.remaining_bit_length == 16
    assert ff.fetch_aligned_u8() == 0b11001101
    assert ff.fetch_aligned_u8() == 0b10101011
    assert ff.remaining_bit_length == 0
    assert ff.consumed_bit_length == 16
    assert ff.fetch_aligned_u8() == 0
    assert ff.fetch_aligned_u16() == 0
    assert ff.fetch_aligned_u32() == 0
    assert ff.fetch_aligned_u64() == 0

    f.skip_bits(40)
    assert f.consumed_bit_length == 56
    assert f.remaining_bit_length == 8
    assert f.fetch_aligned_u8() == 0b00110100
    assert f.remaining_bit_length == 0



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing


T = typing.TypeVar("T")


def mark_last(it: typing.Iterable[T]) -> typing.Iterable[typing.Tuple[bool, T]]:
    """
    This is an iteration helper like :func:`enumerate`. It amends every item with a boolean flag which is False
    for all items except the last one. If the input iterable is empty, yields nothing.

    >>> list(mark_last([]))
    []
    >>> list(mark_last([123]))
    [(True, 123)]
    >>> list(mark_last([123, 456]))
    [(False, 123), (True, 456)]
    >>> list(mark_last([123, 456, 789]))
    [(False, 123), (False, 456), (True, 789)]
    """
    it = iter(it)
    try:
        last = next(it)
    except StopIteration:
        pass
    else:
        for val in it:
            yield False, last
            last = val
        yield True, last



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
Keeps track of online nodes by subscribing to ``uavcan.node.Heartbeat`` and requesting ``uavcan.node.GetInfo``
when necessary; see :class:`pyuavcan.application.node_tracker.NodeTracker`.
"""

from __future__ import annotations
from typing import NamedTuple, Callable, Optional, Dict, List
import asyncio
import logging
from uavcan.node import Heartbeat_1 as Heartbeat
from uavcan.node import GetInfo_1 as GetInfo
import pyuavcan
import pyuavcan.application


__all__ = ["Entry", "UpdateHandler", "NodeTracker"]


Entry = NamedTuple(
    "Entry",
    [
        ("heartbeat", Heartbeat),
        ("info", Optional[GetInfo.Response]),
    ],
)
"""
The data kept per online node.
The heartbeat is the latest received one.
The info is None until the node responds to the GetInfo request.
"""


UpdateHandler = Callable[[int, Optional[Entry], Optional[Entry]], None]
"""
Arguments: node-ID, old entry, new entry. See :meth:`NodeTracker.add_update_handler` for details.
"""


_logger = logging.getLogger(__name__)


class NodeTracker:
    """
    This class is designed for tracking the list of online nodes in real time.
    It subscribes to ``uavcan.node.Heartbeat`` to keep a list of online nodes.
    Whenever a new node appears online or an existing node is restarted
    (restart is detected via the uptime counter),
    the tracker invokes ``uavcan.node.GetInfo`` on it and keeps the response until the node is restarted again
    or until it goes offline (offline nodes detected via heartbeat timeout).
    If the node did not reply to ``uavcan.node.GetInfo``, the request will be retried later.

    If the local node is anonymous, the info request functionality will be automatically disabled;
    it will be re-enabled automatically if the local node is assigned a node-ID later
    (nodes that are already known at this time may not be queried).

    The tracked node registry *does not include the local node*.
    If the local node-ID is N, the registry will not contain an entry at key N unless there is a node-ID conflict
    in the network.

    The class provides IoC events which are triggered on change.
    The collected data can also be accessed by direct polling synchronously.
    """

    GET_INFO_PRIORITY = pyuavcan.transport.Priority.OPTIONAL
    """
    The logic tolerates the loss of responses, hence the optional priority level.
    This way, we can retry without affecting high-priority communications.
    """

    DEFAULT_GET_INFO_TIMEOUT = 5.0
    """
    The default request timeout is larger than the recommended default because the data is immutable
    (does not lose validity over time) and the priority level is low which may cause delays.
    """

    DEFAULT_GET_INFO_ATTEMPTS = 10
    """
    Abandon efforts if the remote node did not respond to GetInfo this many times.
    The counter will resume from scratch if the node is restarted or a new node under that node-ID is detected.
    """

    def __init__(self, node: pyuavcan.application.Node):
        self._node = node
        self._sub_heartbeat = self.node.make_subscriber(Heartbeat)

        self._registry: Dict[int, Entry] = {}
        self._offline_timers: Dict[int, asyncio.TimerHandle] = {}
        self._info_tasks: Dict[int, asyncio.Task[None]] = {}

        self._update_handlers: List[UpdateHandler] = []

        self._get_info_timeout = self.DEFAULT_GET_INFO_TIMEOUT
        self._get_info_attempts = self.DEFAULT_GET_INFO_ATTEMPTS

        def close() -> None:
            """
            When closed the registry is emptied and all handlers are removed.
            This is to avoid accidental reliance on obsolete data.
            """
            _logger.debug("Closing %s", self)
            self._sub_heartbeat.close()
            self._registry.clear()
            self._update_handlers.clear()

            for tm in self._offline_timers.values():
                tm.cancel()
            self._offline_timers.clear()

            for tsk in self._info_tasks.values():
                tsk.cancel()
            self._info_tasks.clear()

        node.add_lifetime_hooks(
            lambda: self._sub_heartbeat.receive_in_background(self._on_heartbeat),
            close,
        )

    @property
    def node(self) -> pyuavcan.application.Node:
        return self._node

    @property
    def get_info_timeout(self) -> float:
        """
        Allows the user to override the default ``uavcan.node.GetInfo`` request timeout.
        The value shall be a finite positive number.
        """
        return self._get_info_timeout

    @get_info_timeout.setter
    def get_info_timeout(self, value: float) -> None:
        value = float(value)
        if 0 < value < float("+inf"):
            self._get_info_timeout = value
        else:
            raise ValueError(f"Invalid response timeout value: {value}")

    @property
    def get_info_attempts(self) -> int:
        """
        Allows the user to override the default ``uavcan.node.GetInfo`` request retry limit.
        The value shall be a non-negative integer number.
        The value of zero disables GetInfo requests completely.
        """
        return self._get_info_attempts

    @get_info_attempts.setter
    def get_info_attempts(self, value: int) -> None:
        value = int(value)
        if 0 <= value:
            self._get_info_attempts = value
        else:
            raise ValueError(f"Invalid attempt limit: {value}")

    @property
    def registry(self) -> Dict[int, Entry]:
        """
        Access the live online node registry. Keys are node-ID, values are :class:`Entry`.
        The returned value is a copy of the actual registry to prevent accidental mutation.
        Elements are ordered by node-ID.
        """
        return {  # pylint: disable=unnecessary-comprehension
            k: v for k, v in sorted(self._registry.items(), key=lambda item: item[0])
        }

    def add_update_handler(self, handler: UpdateHandler) -> None:
        """
        Register a callable that will be invoked whenever the node registry is changed.
        The arguments are: node-ID, old entry, new entry.
        The handler is invoked in the following cases with the specified arguments:

        - New node appeared online. The old-entry is None. The new-entry info is None.
        - A known node went offline. The new-entry is None.
        - A known node restarted. Neither entry is None. The new-entry info is None.
        - A node responds to a ``uavcan.node.GetInfo`` request. Neither entry is None. The new-entry info is not None.

        Received Heartbeat messages change the registry as well, but they do not trigger the hook.
        Handlers can be added and removed at any moment regardless of whether the instance is started.
        """
        if not callable(handler):  # pragma: no cover
            raise ValueError(f"Bad handler: {handler}")
        self._update_handlers.append(handler)

    def remove_update_handler(self, handler: UpdateHandler) -> None:
        """
        Remove a previously added hook identified by referential equivalence. Behaves like :meth:`list.remove`.
        """
        self._update_handlers.remove(handler)

    async def _on_heartbeat(self, msg: Heartbeat, metadata: pyuavcan.transport.TransferFrom) -> None:
        loop = asyncio.get_running_loop()
        node_id = metadata.source_node_id
        if node_id is None:
            _logger.warning("Anonymous nodes shall not publish Heartbeat. Message: %s. Metadata: %s", msg, metadata)
            return

        # Construct the new entry and decide if we need to issue another GetInfo request.
        update = True
        old = self._registry.get(node_id)
        if old is None:
            new = Entry(msg, None)
            _logger.debug("New node %s heartbeat %s", node_id, msg)
        elif old[0].uptime > msg.uptime:
            new = Entry(msg, None)
            _logger.debug("Known node %s restarted. New heartbeat: %s. Old entry: %s", node_id, msg, old)
        else:
            new = Entry(msg, old[1])
            update = False

        # Set up the offline timer that will fire when the Heartbeat messages were not seen for long enough.
        self._registry[node_id] = new
        try:
            self._offline_timers[node_id].cancel()
        except LookupError:
            pass
        self._offline_timers[node_id] = loop.call_later(Heartbeat.OFFLINE_TIMEOUT, self._on_offline, node_id)

        # Do the update unless this is just a regular heartbeat (no restart, known node).
        if update:
            self._request_info(node_id)
            self._notify(node_id, old, new)

    def _on_offline(self, node_id: int) -> None:
        try:
            old = self._registry[node_id]
            _logger.debug("Offline timeout expired for node %s. Old entry: %s", node_id, old)
            self._notify(node_id, old, None)
            del self._registry[node_id]
            self._cancel_task(node_id)
            del self._offline_timers[node_id]
        except Exception as ex:
            _logger.exception("Offline timeout handler error for node %s: %s", node_id, ex)

    def _cancel_task(self, node_id: int) -> None:
        try:
            task = self._info_tasks[node_id]
        except LookupError:
            pass
        else:
            task.cancel()
            del self._info_tasks[node_id]
            _logger.debug("GetInfo task for node %s canceled", node_id)

    def _request_info(self, node_id: int) -> None:
        async def attempt() -> bool:
            client = self.node.make_client(GetInfo, node_id)
            try:
                client.priority = self.GET_INFO_PRIORITY
                client.response_timeout = self._get_info_timeout
                response = await client.call(GetInfo.Request())
                if response is not None:
                    _logger.debug("GetInfo response: %s", response)
                    obj, _meta = response
                    assert isinstance(obj, GetInfo.Response)
                    old = self._registry[node_id]
                    new = Entry(old[0], obj)
                    self._registry[node_id] = new
                    self._notify(node_id, old, new)
                    return True
                _logger.debug("GetInfo request to %s has timed out in %.3f seconds", node_id, client.response_timeout)
                return False
            finally:
                client.close()

        async def worker() -> None:
            try:
                _logger.debug("GetInfo task for node %s started", node_id)
                remaining_attempts = self._get_info_attempts
                while remaining_attempts > 0:
                    _logger.debug(
                        "GetInfo task for node %s is making a new attempt; remaining attempts: %s",
                        node_id,
                        remaining_attempts,
                    )
                    remaining_attempts -= 1
                    try:
                        if await attempt():
                            break
                    except (
                        pyuavcan.transport.OperationNotDefinedForAnonymousNodeError,
                        pyuavcan.presentation.RequestTransferIDVariabilityExhaustedError,
                    ) as ex:
                        _logger.debug("GetInfo task for node %s encountered a transient error: %s", node_id, ex)
                        await asyncio.sleep(self._get_info_timeout)
                _logger.debug("GetInfo task for node %s is exiting", node_id)
            except asyncio.CancelledError:  # pylint: disable=try-except-raise
                raise
            except pyuavcan.transport.ResourceClosedError:
                _logger.debug("GetInfo task for node %s is stopping because the transport is closed.", node_id)
            except Exception as ex:
                _logger.exception("GetInfo task for node %s has crashed: %s", node_id, ex)
            del self._info_tasks[node_id]

        self._cancel_task(node_id)
        self._info_tasks[node_id] = asyncio.get_event_loop().create_task(worker())

    def _notify(self, node_id: int, old_entry: Optional[Entry], new_entry: Optional[Entry]) -> None:
        assert isinstance(old_entry, Entry) or old_entry is None
        assert isinstance(new_entry, Entry) or new_entry is None
        pyuavcan.util.broadcast(self._update_handlers)(node_id, old_entry, new_entry)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
This module is used for automatic generation of Python classes from DSDL type definitions and
also for various manipulations on them.
Auto-generated classes have a high-level application-facing API and built-in auto-generated
serialization and deserialization routines.

The serialization code heavily relies on NumPy and the data alignment analysis implemented in PyDSDL.
Some of the technical details are covered in the following posts:

- https://forum.uavcan.org/t/pyuavcan-design-thread/504
- https://github.com/UAVCAN/pydsdl/pull/24

The main entity of this module is the function :func:`compile`.

Below is the inheritance diagram for the classes defined in this module.

.. inheritance-diagram:: pyuavcan.dsdl._composite_object
   :parts: 1
"""

from ._compiler import compile as compile  # pylint: disable=redefined-builtin
from ._compiler import compile_all as compile_all
from ._compiler import GeneratedPackageInfo as GeneratedPackageInfo

from ._composite_object import serialize as serialize
from ._composite_object import deserialize as deserialize

from ._composite_object import CompositeObject as CompositeObject
from ._composite_object import ServiceObject as ServiceObject

from ._composite_object import FixedPortObject as FixedPortObject
from ._composite_object import FixedPortCompositeObject as FixedPortCompositeObject
from ._composite_object import FixedPortServiceObject as FixedPortServiceObject

from ._composite_object import get_fixed_port_id as get_fixed_port_id
from ._composite_object import get_model as get_model
from ._composite_object import get_class as get_class
from ._composite_object import get_extent_bytes as get_extent_bytes

from ._composite_object import get_attribute as get_attribute
from ._composite_object import set_attribute as set_attribute

from ._builtin_form import to_builtin as to_builtin
from ._builtin_form import update_from_builtin as update_from_builtin


def generate_package(*args, **kwargs):  # type: ignore  # pragma: no cover
    """Deprecated alias of :func:`compile`."""
    import warnings

    warnings.warn(
        "pyuavcan.dsdl.generate_package() is deprecated; use pyuavcan.dsdl.compile() instead.",
        DeprecationWarning,
    )
    return compile(*args, **kwargs)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import copy
import typing
import asyncio
import logging
import dataclasses
import pyuavcan
from pyuavcan.transport import Timestamp
from pyuavcan.transport.commons.high_overhead_transport import TransferReassembler
from .._frame import SerialFrame
from ._base import SerialSession


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SerialInputSessionStatistics(pyuavcan.transport.SessionStatistics):
    reassembly_errors_per_source_node_id: typing.Dict[
        int, typing.Dict[TransferReassembler.Error, int]
    ] = dataclasses.field(default_factory=dict)
    """
    Keys are source node-IDs; values are dicts where keys are error enum members and values are counts.
    """


class SerialInputSession(SerialSession, pyuavcan.transport.InputSession):
    DEFAULT_TRANSFER_ID_TIMEOUT = 2.0
    """
    Units are seconds. Can be overridden after instantiation if needed.
    """

    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly.
        Instead, use the factory method :meth:`pyuavcan.transport.serial.SerialTransport.get_input_session`.
        """
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._statistics = SerialInputSessionStatistics()
        self._transfer_id_timeout = self.DEFAULT_TRANSFER_ID_TIMEOUT
        self._queue: asyncio.Queue[pyuavcan.transport.TransferFrom] = asyncio.Queue()
        self._reassemblers: typing.Dict[int, TransferReassembler] = {}
        super().__init__(finalizer)

    def _process_frame(self, timestamp: Timestamp, frame: SerialFrame) -> None:
        """
        This is a part of the transport-internal API. It's a public method despite the name because Python's
        visibility handling capabilities are limited. I guess we could define a private abstract base to
        handle this but it feels like too much work. Why can't we have protected visibility in Python?
        """
        assert frame.data_specifier == self._specifier.data_specifier, "Internal protocol violation"
        self._statistics.frames += 1

        transfer: typing.Optional[pyuavcan.transport.TransferFrom]
        if frame.source_node_id is None:
            transfer = TransferReassembler.construct_anonymous_transfer(timestamp, frame)
            if transfer is None:
                self._statistics.errors += 1
                _logger.debug("%s: Invalid anonymous frame: %s", self, frame)
        else:
            transfer = self._get_reassembler(frame.source_node_id).process_frame(
                timestamp, frame, self._transfer_id_timeout
            )
        if transfer is not None:
            self._statistics.transfers += 1
            self._statistics.payload_bytes += sum(map(len, transfer.fragmented_payload))
            _logger.debug("%s: Received transfer: %s; current stats: %s", self, transfer, self._statistics)
            try:
                self._queue.put_nowait(transfer)
            except asyncio.QueueFull:  # pragma: no cover
                # TODO: make the queue capacity configurable
                self._statistics.drops += len(transfer.fragmented_payload)

    async def receive(self, monotonic_deadline: float) -> typing.Optional[pyuavcan.transport.TransferFrom]:
        try:
            loop = asyncio.get_running_loop()
            timeout = monotonic_deadline - loop.time()
            if timeout > 0:
                transfer = await asyncio.wait_for(self._queue.get(), timeout)
            else:
                transfer = self._queue.get_nowait()
        except (asyncio.TimeoutError, asyncio.QueueEmpty):
            # If there are unprocessed transfers, allow the caller to read them even if the instance is closed.
            self._raise_if_closed()
            return None
        else:
            assert isinstance(transfer, pyuavcan.transport.TransferFrom), "Internal protocol violation"
            assert transfer.source_node_id == self._specifier.remote_node_id or self._specifier.remote_node_id is None
            return transfer

    @property
    def transfer_id_timeout(self) -> float:
        return self._transfer_id_timeout

    @transfer_id_timeout.setter
    def transfer_id_timeout(self, value: float) -> None:
        if value > 0:
            self._transfer_id_timeout = float(value)
        else:
            raise ValueError(f"Invalid value for transfer-ID timeout [second]: {value}")

    @property
    def specifier(self) -> pyuavcan.transport.InputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> SerialInputSessionStatistics:
        return copy.copy(self._statistics)

    def _get_reassembler(self, source_node_id: int) -> TransferReassembler:
        try:
            return self._reassemblers[source_node_id]
        except LookupError:

            def on_reassembly_error(error: TransferReassembler.Error) -> None:
                self._statistics.errors += 1
                d = self._statistics.reassembly_errors_per_source_node_id[source_node_id]
                try:
                    d[error] += 1
                except LookupError:
                    d[error] = 1

            self._statistics.reassembly_errors_per_source_node_id.setdefault(source_node_id, {})
            reasm = TransferReassembler(
                source_node_id=source_node_id,
                extent_bytes=self._payload_metadata.extent_bytes,
                on_error_callback=on_reassembly_error,
            )
            self._reassemblers[source_node_id] = reasm
            _logger.debug("%s: New %s (%d total)", self, reasm, len(self._reassemblers))
            return reasm



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
UAVCAN/CAN transport overview
+++++++++++++++++++++++++++++

This module implements UAVCAN/CAN -- the CAN transport for UAVCAN, both Classic CAN and CAN FD,
as defined in the UAVCAN specification.
UAVCAN does not distinguish between the two aside from the MTU difference; neither does this implementation.
Classic CAN is essentially treated as CAN FD with MTU of 8 bytes.

Different CAN hardware is supported through the media sublayer; please refer to :mod:`pyuavcan.transport.can.media`.

Per the UAVCAN specification, the CAN transport supports broadcast messages and unicast services:

+--------------------+--------------------------+---------------------------+
| Supported transfers| Unicast                  | Broadcast                 |
+====================+==========================+===========================+
|**Message**         | No                       | Yes                       |
+--------------------+--------------------------+---------------------------+
|**Service**         | Yes                      | Banned by Specification   |
+--------------------+--------------------------+---------------------------+


Tooling
+++++++

Some of the media sub-layer implementations support virtual CAN bus interfaces
(e.g., SocketCAN on GNU/Linux); they are often useful for testing.
Please read the media sub-layer documentation for details.


Inheritance diagram
+++++++++++++++++++

.. inheritance-diagram:: pyuavcan.transport.can._can
                         pyuavcan.transport.can._session._input
                         pyuavcan.transport.can._session._output
                         pyuavcan.transport.can._tracer
   :parts: 1
"""

# Please keep the elements well-ordered because the order is reflected in the docs.
# Core components first.
from ._can import CANTransport as CANTransport

from ._session import CANInputSession as CANInputSession
from ._session import CANOutputSession as CANOutputSession

# Statistics.
from ._can import CANTransportStatistics as CANTransportStatistics

from ._session import CANInputSessionStatistics as CANInputSessionStatistics
from ._session import TransferReassemblyErrorID as TransferReassemblyErrorID

# Analysis.
from ._tracer import CANCapture as CANCapture
from ._tracer import CANErrorTrace as CANErrorTrace
from ._tracer import CANTracer as CANTracer

# Media sub-layer.
from . import media as media



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from ._base import RedundantSession as RedundantSession
from ._base import RedundantSessionStatistics as RedundantSessionStatistics

from ._input import RedundantInputSession as RedundantInputSession
from ._input import RedundantTransferFrom as RedundantTransferFrom

from ._output import RedundantOutputSession as RedundantOutputSession
from ._output import RedundantFeedback as RedundantFeedback



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import logging

R = typing.TypeVar("R")

_logger = logging.getLogger(__name__)


def broadcast(
    functions: typing.Iterable[typing.Callable[..., R]]
) -> typing.Callable[..., typing.List[typing.Union[R, Exception]]]:
    """
    Returns a function that invokes each supplied function in series with the specified arguments
    following the specified order.
    If a function is executed successfully, its result is added to the output list.
    If it raises an exception, the exception is suppressed, logged, and added to the output list instead of the result.

    This function is mostly intended for invoking various handlers.

    ..  doctest::
        :hide:

        >>> _logger.setLevel(100)  # Suppress error reports from the following doctest.

    >>> def add(a, b):
    ...     return a + b
    >>> def fail(a, b):
    ...     raise ValueError(f'Arguments: {a}, {b}')
    >>> broadcast([add, fail])(4, b=5)
    [9, ValueError('Arguments: 4, 5')]
    >>> broadcast([print])('Hello', 'world!')
    Hello world!
    [None]
    >>> broadcast([])()
    []
    """

    def delegate(*args: typing.Any, **kwargs: typing.Any) -> typing.List[typing.Union[R, Exception]]:
        out: typing.List[typing.Union[R, Exception]] = []
        for fn in functions:
            try:
                r: typing.Union[R, Exception] = fn(*args, **kwargs)
            except Exception as ex:
                r = ex
                _logger.exception("Unhandled exception in %s: %s", fn, ex)
            out.append(r)
        return out

    return delegate



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import copy
import typing
import asyncio
import logging
import warnings
import threading
import dataclasses
import concurrent.futures
import serial
import pyuavcan.transport
from pyuavcan.transport import Timestamp
from ._frame import SerialFrame
from ._stream_parser import StreamParser
from ._session import SerialOutputSession, SerialInputSession
from ._tracer import SerialCapture, SerialTracer


_SERIAL_PORT_READ_TIMEOUT = 1.0


_logger = logging.getLogger(__name__)


@dataclasses.dataclass
class SerialTransportStatistics(pyuavcan.transport.TransportStatistics):
    in_bytes: int = 0
    in_frames: int = 0
    in_out_of_band_bytes: int = 0

    out_bytes: int = 0
    out_frames: int = 0
    out_transfers: int = 0
    out_incomplete: int = 0


class SerialTransport(pyuavcan.transport.Transport):
    """
    The UAVCAN/Serial transport is designed for OSI L1 byte-level serial links and tunnels,
    such as UART, RS-422/485/232 (duplex), USB CDC ACM, TCP/IP, etc.
    Please read the module documentation for details.
    """

    TRANSFER_ID_MODULO = SerialFrame.TRANSFER_ID_MASK + 1

    VALID_MTU_RANGE = (1024, 1024 ** 3)
    """
    The maximum MTU is practically unlimited, and it is also the default MTU.
    This is by design to ensure that all frames are single-frame transfers.
    Compliant implementations of the serial transport do not have to support multi-frame transfers,
    which removes the greatest chunk of complexity from the protocol.
    """

    DEFAULT_SERVICE_TRANSFER_MULTIPLIER = 2
    VALID_SERVICE_TRANSFER_MULTIPLIER_RANGE = (1, 5)

    def __init__(
        self,
        serial_port: typing.Union[str, serial.SerialBase],
        local_node_id: typing.Optional[int],
        *,
        mtu: int = max(VALID_MTU_RANGE),
        service_transfer_multiplier: int = DEFAULT_SERVICE_TRANSFER_MULTIPLIER,
        baudrate: typing.Optional[int] = None,
        loop: typing.Optional[asyncio.AbstractEventLoop] = None,
    ):
        """
        :param serial_port: The serial port instance to communicate over, or its name.
            In the latter case, the port will be constructed via :func:`serial.serial_for_url`
            (refer to the PySerial docs for the background).
            The new instance takes ownership of the port; when the instance is closed, its port will also be closed.
            Examples:

            - ``/dev/ttyACM0`` -- a regular serial port on GNU/Linux (USB CDC ACM in this example).
            - ``COM9`` -- likewise, on Windows.
            - ``/dev/serial/by-id/usb-Black_Sphere_Technologies_Black_Magic_Probe_B5DCABF5-if02`` -- a regular
              USB CDC ACM port referenced by the device name and ID (GNU/Linux).
            - ``hwgrep:///dev/serial/by-id/*Black_Magic_Probe*-if02`` -- glob instead of exact name.
            - ``socket://127.0.0.1:50905`` -- a TCP/IP tunnel instead of a physical port.
            - ``spy://COM3?file=dump.txt`` -- open a regular port and dump all data exchange into a text file.

            Read the PySerial docs for more info.

        :param local_node_id: The node-ID to use. Can't be changed after initialization.
            None means that the transport will operate in the anonymous mode.

        :param mtu: Use single-frame transfers for all outgoing transfers containing not more than than
            this many bytes of payload. Otherwise, use multi-frame transfers.

            By default, the MTU is virtually unlimited (to be precise, it is set to a very large number that
            is unattainable in practice), meaning that all transfers will be single-frame transfers.
            Such behavior is optimal for the serial transport because it does not have native framing
            and as such it supports frames of arbitrary sizes. Implementations may omit the support for
            multi-frame transfers completely, which removes the greatest chunk of complexity from the protocol.

            This setting does not affect transfer reception -- the RX MTU is always set to the maximum valid MTU
            (i.e., practically unlimited).

        :param service_transfer_multiplier: Deterministic data loss mitigation for service transfers.
            This parameter specifies the number of times each outgoing service transfer will be repeated.
            This setting does not affect message transfers.

        :param baudrate: If not None, the specified baud rate will be configured on the serial port.
            Otherwise, the baudrate will be left unchanged.

        :param loop: Deprecated.
        """
        self._service_transfer_multiplier = int(service_transfer_multiplier)
        self._mtu = int(mtu)
        if loop:
            warnings.warn("The loop argument is deprecated.", DeprecationWarning)

        low, high = self.VALID_SERVICE_TRANSFER_MULTIPLIER_RANGE
        if not (low <= self._service_transfer_multiplier <= high):
            raise ValueError(f"Invalid service transfer multiplier: {self._service_transfer_multiplier}")

        low, high = self.VALID_MTU_RANGE
        if not (low <= self._mtu <= high):
            raise ValueError(f"Invalid MTU: {self._mtu} bytes")

        self._local_node_id = int(local_node_id) if local_node_id is not None else None
        if self._local_node_id is not None and not (0 <= self._local_node_id < self.protocol_parameters.max_nodes):
            raise ValueError(f"Invalid node ID for serial: {self._local_node_id}")

        # At first I tried using serial.is_open, but unfortunately that doesn't work reliably because the close()
        # method on most serial port classes is non-atomic, which causes all sorts of weird race conditions
        # and spurious errors in the reader thread (at least). A simple explicit flag is reliable.
        self._closed = False

        # For serial port write serialization. Read operations are performed concurrently (no sync) in separate thread.
        self._port_lock = asyncio.Lock()

        # The serialization buffer is re-used for performance reasons; it is needed to store frame contents before
        # they are emitted into the serial port. It may grow as necessary at runtime; the initial size is a guess.
        # Access must be protected with the port lock!
        self._serialization_buffer = bytearray(b"\x00" * (1024 * 1024))

        self._input_registry: typing.Dict[pyuavcan.transport.InputSessionSpecifier, SerialInputSession] = {}
        self._output_registry: typing.Dict[pyuavcan.transport.OutputSessionSpecifier, SerialOutputSession] = {}

        self._capture_handlers: typing.List[pyuavcan.transport.CaptureCallback] = []

        self._statistics = SerialTransportStatistics()

        if not isinstance(serial_port, serial.SerialBase):
            serial_port = serial.serial_for_url(serial_port)
        assert isinstance(serial_port, serial.SerialBase)
        if not serial_port.is_open:
            raise pyuavcan.transport.InvalidMediaConfigurationError("The serial port instance is not open")
        serial_port.timeout = _SERIAL_PORT_READ_TIMEOUT
        self._serial_port = serial_port
        if baudrate is not None:
            self._serial_port.baudrate = int(baudrate)

        self._background_executor = concurrent.futures.ThreadPoolExecutor()

        self._reader_thread = threading.Thread(
            target=self._reader_thread_func, args=(asyncio.get_event_loop(),), daemon=True
        )
        self._reader_thread.start()

    @property
    def protocol_parameters(self) -> pyuavcan.transport.ProtocolParameters:
        return pyuavcan.transport.ProtocolParameters(
            transfer_id_modulo=self.TRANSFER_ID_MODULO,
            max_nodes=len(SerialFrame.NODE_ID_RANGE),
            mtu=self._mtu,
        )

    @property
    def local_node_id(self) -> typing.Optional[int]:
        return self._local_node_id

    def close(self) -> None:
        self._closed = True
        for s in (*self.input_sessions, *self.output_sessions):
            try:
                s.close()
            except Exception as ex:  # pragma: no cover
                _logger.exception("%s: Failed to close session %r: %s", self, s, ex)

        if self._serial_port.is_open:  # Double-close is not an error.
            self._serial_port.close()

    def get_input_session(
        self, specifier: pyuavcan.transport.InputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> SerialInputSession:
        def finalizer() -> None:
            del self._input_registry[specifier]

        self._ensure_not_closed()
        try:
            out = self._input_registry[specifier]
        except LookupError:
            out = SerialInputSession(specifier=specifier, payload_metadata=payload_metadata, finalizer=finalizer)
            self._input_registry[specifier] = out

        assert isinstance(out, SerialInputSession)
        assert specifier in self._input_registry
        assert out.specifier == specifier
        return out

    def get_output_session(
        self, specifier: pyuavcan.transport.OutputSessionSpecifier, payload_metadata: pyuavcan.transport.PayloadMetadata
    ) -> SerialOutputSession:
        self._ensure_not_closed()
        if specifier not in self._output_registry:

            def finalizer() -> None:
                del self._output_registry[specifier]

            if (
                isinstance(specifier.data_specifier, pyuavcan.transport.ServiceDataSpecifier)
                and self._service_transfer_multiplier > 1
            ):

                async def send_transfer(
                    frames: typing.List[SerialFrame], monotonic_deadline: float
                ) -> typing.Optional[Timestamp]:
                    first_tx_ts: typing.Optional[Timestamp] = None
                    for _ in range(self._service_transfer_multiplier):  # pragma: no branch
                        ts = await self._send_transfer(frames, monotonic_deadline)
                        first_tx_ts = first_tx_ts or ts
                    return first_tx_ts

            else:
                send_transfer = self._send_transfer

            self._output_registry[specifier] = SerialOutputSession(
                specifier=specifier,
                payload_metadata=payload_metadata,
                mtu=self._mtu,
                local_node_id=self._local_node_id,
                send_handler=send_transfer,
                finalizer=finalizer,
            )

        out = self._output_registry[specifier]
        assert isinstance(out, SerialOutputSession)
        assert out.specifier == specifier
        return out

    @property
    def input_sessions(self) -> typing.Sequence[SerialInputSession]:
        return list(self._input_registry.values())

    @property
    def output_sessions(self) -> typing.Sequence[SerialOutputSession]:
        return list(self._output_registry.values())

    @property
    def serial_port(self) -> serial.SerialBase:
        assert isinstance(self._serial_port, serial.SerialBase)
        return self._serial_port

    def sample_statistics(self) -> SerialTransportStatistics:
        return copy.copy(self._statistics)

    def begin_capture(self, handler: pyuavcan.transport.CaptureCallback) -> None:
        """
        The reported events are of type :class:`SerialCapture`, please read its documentation for details.
        The events may be reported from a different thread (use locks).
        """
        self._capture_handlers.append(handler)

    @property
    def capture_active(self) -> bool:
        return len(self._capture_handlers) > 0

    @staticmethod
    def make_tracer() -> SerialTracer:
        """
        See :class:`SerialTracer`.
        """
        return SerialTracer()

    async def spoof(self, transfer: pyuavcan.transport.AlienTransfer, monotonic_deadline: float) -> bool:
        """
        Spoofing over the serial transport is trivial and it does not involve reconfiguration of the media layer.
        It can be invoked at no cost at any time (unlike, say, UAVCAN/UDP).
        See the overridden method :meth:`pyuavcan.transport.Transport.spoof` for details.

        Notice that if the transport operates over the virtual loopback port ``loop://`` with capture enabled,
        every spoofed frame will be captured twice: one TX, one RX. Same goes for regular transfers.
        """

        ss = transfer.metadata.session_specifier
        src, dst = ss.source_node_id, ss.destination_node_id
        if isinstance(ss.data_specifier, pyuavcan.transport.ServiceDataSpecifier) and (src is None or dst is None):
            raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                f"Anonymous nodes cannot participate in service calls. Spoof metadata: {transfer.metadata}"
            )

        def construct_frame(index: int, end_of_transfer: bool, payload: memoryview) -> SerialFrame:
            if not end_of_transfer and src is None:
                raise pyuavcan.transport.OperationNotDefinedForAnonymousNodeError(
                    f"Anonymous nodes cannot emit multi-frame transfers. Spoof metadata: {transfer.metadata}"
                )
            return SerialFrame(
                priority=transfer.metadata.priority,
                transfer_id=transfer.metadata.transfer_id,
                index=index,
                end_of_transfer=end_of_transfer,
                payload=payload,
                source_node_id=src,
                destination_node_id=dst,
                data_specifier=ss.data_specifier,
            )

        frames = list(
            pyuavcan.transport.commons.high_overhead_transport.serialize_transfer(
                transfer.fragmented_payload, self._mtu, construct_frame
            )
        )
        _logger.debug("%s: Spoofing %s", self, frames)
        return await self._send_transfer(frames, monotonic_deadline) is not None

    def _handle_received_frame(self, timestamp: Timestamp, frame: SerialFrame) -> None:
        self._statistics.in_frames += 1
        if frame.destination_node_id in (self._local_node_id, None):
            for source_node_id in {None, frame.source_node_id}:  # pylint: disable=use-sequence-for-iteration
                ss = pyuavcan.transport.InputSessionSpecifier(frame.data_specifier, source_node_id)
                try:
                    session = self._input_registry[ss]
                except LookupError:
                    pass
                else:
                    session._process_frame(timestamp, frame)  # pylint: disable=protected-access

    def _handle_received_out_of_band_data(self, timestamp: Timestamp, data: memoryview) -> None:
        self._statistics.in_out_of_band_bytes += len(data)
        printable: typing.Union[str, bytes] = bytes(data)
        try:
            assert isinstance(printable, bytes)
            printable = printable.decode("utf8")
        except ValueError:
            pass
        _logger.warning("%s: Out-of-band received at %s: %r", self._serial_port.name, timestamp, printable)

    def _handle_received_item_and_update_stats(
        self, timestamp: Timestamp, item: typing.Union[SerialFrame, memoryview], in_bytes_count: int
    ) -> None:
        if isinstance(item, SerialFrame):
            self._handle_received_frame(timestamp, item)
        elif isinstance(item, memoryview):
            self._handle_received_out_of_band_data(timestamp, item)
        else:
            assert False

        assert self._statistics.in_bytes <= in_bytes_count
        self._statistics.in_bytes = int(in_bytes_count)

    async def _send_transfer(
        self, frames: typing.List[SerialFrame], monotonic_deadline: float
    ) -> typing.Optional[Timestamp]:
        """
        Emits the frames belonging to the same transfer, returns the first frame transmission timestamp.
        The returned timestamp can be used for transfer feedback implementation.
        Aborts if the frames cannot be emitted before the deadline or if a write call fails.
        :returns: The first frame transmission timestamp if all frames are sent successfully.
            None on timeout or on write failure.
        """
        tx_ts: typing.Optional[Timestamp] = None
        self._ensure_not_closed()
        loop = asyncio.get_running_loop()
        try:  # Jeez this is getting complex
            num_sent = 0
            for fr in frames:
                async with self._port_lock:  # TODO: the lock acquisition should be prioritized by frame priority!
                    min_buffer_size = len(fr.payload) * 2
                    if len(self._serialization_buffer) < min_buffer_size:
                        _logger.debug(
                            "%s: The serialization buffer is being enlarged from %d to %d bytes",
                            self,
                            len(self._serialization_buffer),
                            min_buffer_size,
                        )
                        self._serialization_buffer = bytearray(0 for _ in range(min_buffer_size))
                    compiled = fr.compile_into(self._serialization_buffer)
                    timeout = monotonic_deadline - loop.time()
                    if timeout > 0:
                        self._serial_port.write_timeout = timeout
                        try:
                            num_written = await loop.run_in_executor(
                                self._background_executor, self._serial_port.write, compiled
                            )
                            tx_ts = tx_ts or Timestamp.now()
                        except serial.SerialTimeoutException:
                            num_written = 0
                            _logger.info("%s: Port write timed out in %.3fs on frame %r", self, timeout, fr)
                        else:
                            if self._capture_handlers:  # Create a copy to decouple data from the serialization buffer!
                                cap = SerialCapture(tx_ts, memoryview(bytes(compiled)), own=True)
                                pyuavcan.util.broadcast(self._capture_handlers)(cap)
                        self._statistics.out_bytes += num_written or 0
                    else:
                        tx_ts = None  # Timed out
                        break

                num_written = len(compiled) if num_written is None else num_written
                if num_written < len(compiled):
                    tx_ts = None  # Write failed
                    break
                num_sent += 1

            self._statistics.out_frames += num_sent
        except Exception as ex:
            if self._closed:
                raise pyuavcan.transport.ResourceClosedError(f"{self} is closed, transmission aborted.") from ex
            raise
        else:
            if tx_ts is not None:
                self._statistics.out_transfers += 1
            else:
                self._statistics.out_incomplete += 1
            return tx_ts

    def _reader_thread_func(self, loop: asyncio.AbstractEventLoop) -> None:
        in_bytes_count = 0

        def callback(ts: Timestamp, buf: memoryview, frame: typing.Optional[SerialFrame]) -> None:
            item = buf if frame is None else frame
            loop.call_soon_threadsafe(self._handle_received_item_and_update_stats, ts, item, in_bytes_count)
            if self._capture_handlers:
                pyuavcan.util.broadcast(self._capture_handlers)(SerialCapture(ts, buf, own=False))

        try:
            parser = StreamParser(callback, max(self.VALID_MTU_RANGE))
            assert abs(self._serial_port.timeout - _SERIAL_PORT_READ_TIMEOUT) < 0.1

            while not self._closed and self._serial_port.is_open:
                chunk = self._serial_port.read(max(1, self._serial_port.inWaiting()))
                chunk_ts = Timestamp.now()
                in_bytes_count += len(chunk)
                parser.process_next_chunk(chunk, chunk_ts)

        except Exception as ex:  # pragma: no cover
            if self._closed or not self._serial_port.is_open:
                _logger.debug("%s: The serial port is closed, exception ignored: %r", self, ex)
            else:
                _logger.exception(
                    "%s: Reader thread has failed, the instance with port %s will be terminated: %s",
                    self,
                    self._serial_port,
                    ex,
                )
            self._closed = True
            self._serial_port.close()

        finally:
            _logger.debug("%s: Reader thread is exiting. Head aega.", self)

    def _ensure_not_closed(self) -> None:
        if self._closed:
            raise pyuavcan.transport.ResourceClosedError(f"{self} is closed")

    def _get_repr_fields(self) -> typing.Tuple[typing.List[typing.Any], typing.Dict[str, typing.Any]]:
        kwargs = {
            "local_node_id": self.local_node_id,
            "service_transfer_multiplier": self._service_transfer_multiplier,
            "baudrate": self._serial_port.baudrate,
        }
        if self._mtu < max(SerialTransport.VALID_MTU_RANGE):
            kwargs["mtu"] = self._mtu
        return [repr(self._serial_port.name)], kwargs



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import dataclasses
import pyuavcan.transport.loopback
from pyuavcan.transport import Trace, TransferTrace, Capture


@dataclasses.dataclass(frozen=True)
class LoopbackCapture(pyuavcan.transport.Capture):
    """
    Since the loopback transport is not really a transport, its capture events contain entire transfers.
    """

    transfer: pyuavcan.transport.AlienTransfer

    @staticmethod
    def get_transport_type() -> typing.Type[pyuavcan.transport.loopback.LoopbackTransport]:
        return pyuavcan.transport.loopback.LoopbackTransport


class LoopbackTracer(pyuavcan.transport.Tracer):
    """
    Since loopback transport does not have frames to trace, this tracer simply returns the transfer
    from the capture object.
    """

    def update(self, cap: Capture) -> typing.Optional[Trace]:
        if isinstance(cap, LoopbackCapture):
            return TransferTrace(cap.timestamp, cap.transfer, transfer_id_timeout=0)
        return None



# Copyright (c) 2021 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import os
from typing import Callable, Optional, Union, List, Dict
from pathlib import Path
import logging
from . import register


EnvironmentVariables = Union[Dict[str, bytes], Dict[str, str], Dict[bytes, bytes]]


class SimpleRegistry(register.Registry):
    def __init__(
        self,
        register_file: Union[None, str, Path] = None,
        environment_variables: Optional[EnvironmentVariables] = None,
    ) -> None:
        from .register.backend.dynamic import DynamicBackend
        from .register.backend.static import StaticBackend

        self._backend_static = StaticBackend(register_file)
        self._backend_dynamic = DynamicBackend()

        if environment_variables is None:
            try:
                environment_variables = os.environb  # type: ignore
            except AttributeError:  # pragma: no cover
                environment_variables = os.environ  # type: ignore

        assert environment_variables is not None
        self._environment_variables: Dict[str, bytes] = {
            (k if isinstance(k, str) else k.decode()): (v if isinstance(v, bytes) else v.encode())
            for k, v in environment_variables.items()
        }
        super().__init__()

        self._update_from_environment_variables()

    @property
    def backends(self) -> List[register.backend.Backend]:
        return [self._backend_static, self._backend_dynamic]

    @property
    def environment_variables(self) -> Dict[str, bytes]:
        return self._environment_variables

    def _create_static(self, name: str, value: register.Value) -> None:
        _logger.debug("%r: Create static %r = %r", self, name, value)
        self._backend_static[name] = value

    def _create_dynamic(
        self,
        name: str,
        getter: Callable[[], register.Value],
        setter: Optional[Callable[[register.Value], None]],
    ) -> None:
        _logger.debug("%r: Create dynamic %r from getter=%r setter=%r", self, name, getter, setter)
        self._backend_dynamic[name] = getter if setter is None else (getter, setter)

    def _update_from_environment_variables(self) -> None:
        for name in self:
            env_val = self.environment_variables.get(register.get_environment_variable_name(name))
            if env_val is not None:
                _logger.debug("Updating register %r from env: %r", name, env_val)
                reg_val = self[name]
                reg_val.assign_environment_variable(env_val)
                self[name] = reg_val


def make_registry(
    register_file: Union[None, str, Path] = None,
    environment_variables: Optional[EnvironmentVariables] = None,
) -> register.Registry:
    """
    Construct a new instance of :class:`pyuavcan.application.register.Registry`.
    Complex applications with uncommon requirements may choose to implement Registry manually
    instead of using this factory.

    See also: standard RPC-service ``uavcan.register.Access``.

    :param register_file:
        Path to the registry file; or, in other words, the configuration file of this application/node.
        If not provided (default), the registers of this instance will be stored in-memory (volatile configuration).
        If path is provided but the file does not exist, it will be created automatically.
        See :attr:`Node.registry`.

    :param environment_variables:
        During initialization, all registers will be updated based on the environment variables passed here.
        This dict is used to initialize :attr:`pyuavcan.application.register.Registry.environment_variables`.
        Registers that are created later using :meth:`pyuavcan.application.register.Registry.setdefault`
        will use these values as well.

        If None (which is default), the value is initialized by copying :data:`os.environb`.
        Pass an empty dict here to disable environment variable processing.

    :raises:
        - :class:`pyuavcan.application.register.ValueConversionError` if a register is found but its value
          cannot be converted to the correct type, or if the value of an environment variable for a register
          is invalid or incompatible with the register's type
          (e.g., an environment variable set to ``Hello world`` cannot be assigned to register of type ``real64[3]``).
    """
    return SimpleRegistry(register_file, environment_variables)


_logger = logging.getLogger(__name__)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>


class TransportError(RuntimeError):
    """
    This is the root exception class for all transport-related errors.
    Exception types defined at the higher layers up the protocol stack (e.g., the presentation layer)
    also inherit from this type, so the application may use this type as the base exception type for all
    UAVCAN-related errors that occur at runtime.

    This exception type hierarchy is intentionally separated from DSDL-related errors that may occur at
    code generation time.
    """


class InvalidTransportConfigurationError(TransportError):
    """
    The transport could not be initialized or the operation could not be performed
    because the specified configuration is invalid.
    """


class InvalidMediaConfigurationError(InvalidTransportConfigurationError):
    """
    The transport could not be initialized or the operation could not be performed
    because the specified media configuration is invalid.
    """


class UnsupportedSessionConfigurationError(TransportError):
    """
    The requested session configuration is not supported by this transport.
    For example, this exception would be raised if one attempted to create a unicast output for messages over
    the CAN bus transport.
    """


class OperationNotDefinedForAnonymousNodeError(TransportError):
    """
    The requested action would normally be possible, but it is currently not because the transport instance does not
    have a node-ID assigned.
    """


class ResourceClosedError(TransportError):
    """
    The requested operation could not be performed because an associated resource has already been terminated.
    Double-close should not raise exceptions.
    """



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import dataclasses


@dataclasses.dataclass(frozen=True)
class PayloadMetadata:
    """
    This information is obtained from the data type definition.

    Eventually, this type might include the runtime type identification information,
    if it is ever implemented in UAVCAN. The alpha revision used to contain the "data type hash" field here,
    but this concept was found deficient and removed from the proposal.
    You can find related discussion in https://forum.uavcan.org/t/alternative-transport-protocols-in-uavcan/324.
    """

    extent_bytes: int
    """
    The minimum amount of memory required to hold any serialized representation of any compatible version
    of the data type; or, on other words, it is the the maximum possible size of received objects.
    The size is specified in bytes because extent is guaranteed (by definition) to be an integer number of bytes long.

    This parameter is determined by the data type author at the data type definition time.
    It is typically larger than the maximum object size in order to allow the data type author to
    introduce more fields in the future versions of the type;
    for example, ``MyMessage.1.0`` may have the maximum size of 100 bytes and the extent 200 bytes;
    a revised version ``MyMessage.1.1`` may have the maximum size anywhere between 0 and 200 bytes.
    It is always safe to pick a larger value if not sure.
    You will find a more rigorous description in the UAVCAN Specification.

    Transport implementations may use this information to statically size receive buffers.
    """

    def __post_init__(self) -> None:
        if self.extent_bytes < 0:
            raise ValueError(f"Invalid extent [byte]: {self.extent_bytes}")



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import time
import socket
import typing
import select
import asyncio
import logging
import threading
import ipaddress
import functools
import dataclasses
import pyuavcan
from pyuavcan.transport import Timestamp
from ._frame import UDPFrame
from ._ip import unicast_ip_to_node_id


_READ_SIZE = 0xFFFF  # Per libpcap documentation, this is to be sufficient always.
_READ_TIMEOUT = 1.0

_logger = logging.getLogger(__name__)

_parse_address = functools.lru_cache(None)(ipaddress.ip_address)

_IPAddress = typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]


@dataclasses.dataclass
class SocketReaderStatistics:
    """
    Incoming UDP datagram statistics for an input socket.
    """

    accepted_datagrams: typing.Dict[int, int] = dataclasses.field(default_factory=dict)
    """
    Key is the remote node-ID; value is the number of datagrams received from that node.
    The counters are invariant to the validity of the frame contained in the datagram.
    """

    dropped_datagrams: typing.Dict[typing.Union[_IPAddress, int], int] = dataclasses.field(default_factory=dict)
    """
    Counters of datagrams received from IP addresses that cannot be mapped to a valid node-ID,
    and from nodes that no listener is registered for.
    In the former case, the key is the IP address; in the latter case, the key is the node-ID.
    The counters are invariant to the validity of the frame contained in the datagram.
    """


class SocketReader:
    """
    This class is the solution to the UDP demultiplexing problem.
    The objective is to read data from the supplied socket, parse it, and then forward it to interested listeners.

    Why can't we ask the operating system to do this for us? Because there is no portable way of doing this
    (except for multicast sockets).
    Even on GNU/Linux, there is a risk of race conditions, but I'll spare you the details.
    Those who care may read this: https://stackoverflow.com/a/54156768/1007777.

    The UDP transport is unable to detect a node-ID conflict because it has to discard traffic generated
    by itself in user space. To this transport, its own traffic and a node-ID conflict would look identical.
    """

    Listener = typing.Callable[[Timestamp, int, typing.Optional[UDPFrame]], None]
    """
    The callback is invoked with the timestamp, source node-ID, and the frame instance upon successful reception.
    Remember that on UDP there is no concept of "anonymous node", there is DHCP to handle that.
    If a UDP frame is received that does not contain a valid UAVCAN frame,
    the callback is invoked with None for error statistic collection purposes.
    """

    def __init__(
        self,
        sock: socket.socket,
        local_ip_address: _IPAddress,
        anonymous: bool,
        statistics: SocketReaderStatistics,
    ):
        """
        :param sock: The instance takes ownership of the socket; it will be closed when the instance is closed.
        :param local_ip_address: Needed for node-ID mapping.
        :param anonymous: If True, then packets originating from the local IP address will not be discarded.
        :param statistics: A reference to the external statistics object that will be updated by the instance.
        """
        self._sock = sock
        self._sock.setblocking(False)
        self._original_file_desc = self._sock.fileno()  # This is needed for repr() only.
        self._local_ip_address = local_ip_address
        self._anonymous = anonymous
        self._statistics = statistics

        assert isinstance(self._local_ip_address, (ipaddress.IPv4Address, ipaddress.IPv6Address))
        assert isinstance(self._anonymous, bool)
        assert isinstance(self._statistics, SocketReaderStatistics)

        self._listeners: typing.Dict[typing.Optional[int], SocketReader.Listener] = {}
        self._ctl_worker, self._ctl_main = socket.socketpair()  # For communicating with the worker thread.
        self._thread = threading.Thread(
            target=self._thread_entry_point,
            args=(asyncio.get_event_loop(),),
            name=f"socket_reader_fd_{self._original_file_desc}",
            daemon=True,
        )
        self._thread.start()

    def add_listener(self, source_node_id: typing.Optional[int], handler: Listener) -> None:
        """
        :param source_node_id: The listener will be invoked whenever a frame from this node-ID is received.
            If the value is None, the listener will be invoked for all source node-IDs (promiscuous).
            There shall be at most one listener per source node-ID value (incl. None, i.e., at most one
            promiscuous listener).
            If such listener already exists, a :class:`ValueError` will be raised.

        :param handler: The callable of type :attr:`Listener` that received frames will be passed to.
            If a frame is received that cannot be parsed, the callable will be invoked with None
            in order to let it update its error statistics.
        """
        if not self._thread.is_alive():
            raise pyuavcan.transport.ResourceClosedError(f"{self} is no longer operational")

        if source_node_id in self._listeners:
            raise ValueError(
                f"{self}: The listener for node-ID {source_node_id} is already registered "
                f"with handler {self._listeners[source_node_id]}"
            )
        _logger.debug(
            "%r: Adding listener %r for node-ID %r. Current stats: %s", self, handler, source_node_id, self._statistics
        )
        self._listeners[source_node_id] = handler

    def remove_listener(self, node_id: typing.Optional[int]) -> None:
        """
        Raises :class:`LookupError` if there is no such listener.
        """
        _logger.debug("%r: Removing listener for node-ID %r. Current stats: %s", self, node_id, self._statistics)
        del self._listeners[node_id]

    @property
    def has_listeners(self) -> bool:
        """
        If there are no listeners, the reader instance can be safely closed and destroyed.
        """
        return len(self._listeners) > 0

    def close(self) -> None:
        """
        Closes the instance and its socket, waits for the thread to terminate (which should happen instantly).

        This method is guaranteed to not return until the socket is closed and all calls that might have been
        blocked on it have been completed (particularly, the calls made by the worker thread).
        THIS IS EXTREMELY IMPORTANT because if the worker thread is left on a blocking read from a closed socket,
        the next created socket is likely to receive the same file descriptor and the worker thread would then
        inadvertently consume the data destined for another reader.
        Worse yet, this error may occur spuriously depending on the timing of the worker thread's access to the
        blocking read function, causing the problem to appear and disappear at random.
        I literally spent the whole day sifting through logs and Wireshark dumps trying to understand why the test
        (specifically, the node tracker test, which is an application-layer entity)
        sometimes fails to see a service response that is actually present on the wire.
        This case is now covered by a dedicated unit test.

        The lesson is to never close a file descriptor while there is a system call blocked on it. Never again.

        Once closed, new listeners can no longer be added.
        Raises :class:`RuntimeError` instead of closing if there is at least one active listener.
        """
        if self.has_listeners:
            raise RuntimeError("Refusing to close socket reader with active listeners. Call remove_listener first.")
        if self._sock.fileno() < 0:  # Ensure idempotency.
            return
        started_at = time.monotonic()
        try:
            _logger.debug("%r: Stopping the thread before closing the socket to avoid accidental fd reuse...", self)
            self._ctl_main.send(b"stop")  # The actual data is irrelevant, we just need it to unblock the select().
            self._thread.join(timeout=_READ_TIMEOUT)
        finally:
            self._sock.close()
            self._ctl_worker.close()
            self._ctl_main.close()
        _logger.debug("%r: Closed. Elapsed time: %.3f milliseconds", self, (time.monotonic() - started_at) * 1e3)

    def _dispatch_frame(
        self, timestamp: Timestamp, source_ip_address: _IPAddress, frame: typing.Optional[UDPFrame]
    ) -> None:
        # Do not accept datagrams emitted by the local node itself. Do not update the statistics either.
        external = self._anonymous or (source_ip_address != self._local_ip_address)
        if not external:
            return

        # Process the datagram. This is where the actual demultiplexing takes place.
        # The node-ID mapper will return None for datagrams coming from outside of our UAVCAN subnet.
        handled = False
        source_node_id = unicast_ip_to_node_id(self._local_ip_address, source_ip_address)
        if source_node_id is not None:
            # Each frame is sent to the promiscuous listener and to the selective listener.
            # We parse the frame before invoking the listener in order to avoid the double parsing workload.
            for key in (None, source_node_id):
                try:
                    callback = self._listeners[key]
                except LookupError:
                    pass
                else:
                    handled = True
                    try:
                        callback(timestamp, source_node_id, frame)
                    except Exception as ex:  # pragma: no cover
                        _logger.exception("%r: Unhandled exception in the listener for node-ID %r: %s", self, key, ex)

        # Update the statistics.
        if not handled:
            ip_nid: typing.Union[_IPAddress, int] = source_node_id if source_node_id is not None else source_ip_address
            try:
                self._statistics.dropped_datagrams[ip_nid] += 1
            except LookupError:
                self._statistics.dropped_datagrams[ip_nid] = 1
        else:
            assert source_node_id is not None
            try:
                self._statistics.accepted_datagrams[source_node_id] += 1
            except LookupError:
                self._statistics.accepted_datagrams[source_node_id] = 1

    def _thread_entry_point(self, loop: asyncio.AbstractEventLoop) -> None:
        while self._sock.fileno() >= 0:
            try:
                read_ready, _, _ = select.select([self._ctl_worker, self._sock], [], [], _READ_TIMEOUT)
                if self._sock in read_ready:
                    # TODO: use socket timestamping when running on GNU/Linux (Windows does not support timestamping).
                    ts = pyuavcan.transport.Timestamp.now()

                    # Notice that we MUST create a new buffer for each received datagram to avoid race conditions.
                    # Buffer memory cannot be shared because the rest of the stack is completely zero-copy;
                    # meaning that the data we allocate here, at the very bottom of the protocol stack,
                    # is likely to be carried all the way up to the application layer without being copied.
                    data, endpoint = self._sock.recvfrom(_READ_SIZE)
                    assert len(data) < _READ_SIZE, "Datagram might have been truncated"
                    source_ip = _parse_address(endpoint[0])

                    frame = UDPFrame.parse(memoryview(data))
                    _logger.debug(
                        "%r: Received UDP packet of %d bytes from %s containing frame: %s",
                        self,
                        len(data),
                        endpoint,
                        frame,
                    )
                    loop.call_soon_threadsafe(self._dispatch_frame, ts, source_ip, frame)

                if self._ctl_worker in read_ready:
                    cmd = self._ctl_worker.recv(_READ_SIZE)
                    if cmd:
                        _logger.debug("%r: Worker thread has received the stop signal: %r", self, cmd)
                        break
            except Exception as ex:  # pragma: no cover
                _logger.exception("%r: Worker thread error: %s; will continue after a short nap", self, ex)
                time.sleep(1)
        _logger.debug("%r: Worker thread is exiting, bye bye", self)

    def __repr__(self) -> str:
        """
        The instance remembers its original file descriptor and prints it for diagnostic purposes even if the socket
        is already closed.
        Also, the object ID is printed to differentiate instances sharing the same file descriptor
        (which is NOT permitted, of course, but there have been issues related to that so it was added for debugging).
        """
        return pyuavcan.util.repr_attributes_noexcept(
            self,
            id=hex(id(self)),
            original_fd=self._original_file_desc,
            socket=self._sock,
            remote_node_ids=list(self._listeners.keys()),
        )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import asyncio
import logging
import dataclasses
import pyuavcan.transport
import pyuavcan.util
from ._base import RedundantSession, RedundantSessionStatistics
from .._deduplicator import Deduplicator


_logger = logging.getLogger(__name__)


@dataclasses.dataclass(frozen=True, repr=False)
class RedundantTransferFrom(pyuavcan.transport.TransferFrom):
    inferior_session: pyuavcan.transport.InputSession


@dataclasses.dataclass(frozen=True)
class _Inferior:
    session: pyuavcan.transport.InputSession
    worker: asyncio.Task[None]

    def close(self) -> None:
        try:
            self.session.close()
        finally:
            self.worker.cancel()

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self, session=self.session, iface_id=f"{id(self.session):016x}", worker=self.worker
        )


class RedundantInputSession(RedundantSession, pyuavcan.transport.InputSession):
    """
    This is a composite of a group of :class:`pyuavcan.transport.InputSession`.

    The transfer deduplication strategy is chosen between cyclic and monotonic automatically
    when the first inferior is added.
    """

    _READ_TIMEOUT = 1.0

    def __init__(
        self,
        specifier: pyuavcan.transport.InputSessionSpecifier,
        payload_metadata: pyuavcan.transport.PayloadMetadata,
        tid_modulo_provider: typing.Callable[[], int],
        finalizer: typing.Callable[[], None],
    ):
        """
        Do not call this directly! Use the factory method instead.
        """
        self._specifier = specifier
        self._payload_metadata = payload_metadata
        self._get_tid_modulo = tid_modulo_provider
        self._finalizer: typing.Optional[typing.Callable[[], None]] = finalizer
        assert isinstance(self._specifier, pyuavcan.transport.InputSessionSpecifier)
        assert isinstance(self._payload_metadata, pyuavcan.transport.PayloadMetadata)
        assert isinstance(self._get_tid_modulo(), (type(None), int))
        assert callable(self._finalizer)

        self._inferiors: typing.List[_Inferior] = []
        self._deduplicator: typing.Optional[Deduplicator] = None

        # The actual deduplicated transfers received by the inferiors.
        self._read_queue: asyncio.Queue[RedundantTransferFrom] = asyncio.Queue()
        # Queuing errors is meaningless because they lose relevance immediately, so the queue is only one item deep.
        self._error_queue: asyncio.Queue[Exception] = asyncio.Queue(1)

        self._stat_transfers = 0
        self._stat_payload_bytes = 0
        self._stat_errors = 0

    def _add_inferior(self, session: pyuavcan.transport.Session) -> None:
        assert isinstance(session, pyuavcan.transport.InputSession)
        assert self._finalizer is not None, "The session was supposed to be unregistered"
        assert session.specifier == self.specifier and session.payload_metadata == self.payload_metadata
        if session in self.inferiors:
            return
        _logger.debug("%s: Adding inferior %s id=%016x", self, session, id(session))

        # Ensure that the deduplicator is constructed when the first inferior is launched.
        if self._deduplicator is None:
            self._deduplicator = Deduplicator.new(self._get_tid_modulo())
            _logger.debug("%s: Constructed new deduplicator: %s", self, self._deduplicator)

        # Synchronize the settings for the newly added inferior with its siblings.
        # If there are no other inferiors, the first added one seeds the configuration for its future siblings.
        if self._inferiors:
            session.transfer_id_timeout = self.transfer_id_timeout

        # Launch the inferior's worker task in the last order and add that to the registry.
        task = asyncio.get_event_loop().create_task(self._inferior_worker_task(session))
        self._inferiors.append(_Inferior(session=session, worker=task))

    def _close_inferior(self, session_index: int) -> None:
        assert session_index >= 0, "Negative indexes may lead to unexpected side effects"
        assert self._finalizer is not None, "The session was supposed to be unregistered"
        try:
            inf = self._inferiors.pop(session_index)
        except LookupError:
            pass
        else:
            _logger.debug(
                "%s: Closing inferior %s that used to reside at index %d. Remaining siblings: %s",
                self,
                inf,
                session_index,
                self._inferiors,
            )
            inf.close()
        finally:
            if not self._inferiors:
                # Reset because inferiors we add later may require a different deduplication strategy.
                # When no inferiors are left, there are no consistency constraints to respect.
                self._deduplicator = None

    @property
    def inferiors(self) -> typing.Sequence[pyuavcan.transport.InputSession]:
        return [x.session for x in self._inferiors]

    async def receive(self, monotonic_deadline: float) -> typing.Optional[RedundantTransferFrom]:
        """
        Reads one deduplicated transfer received from all inferiors concurrently. Returns None on timeout.
        If there are no inferiors at the time of the invocation and none appear by the expiration of the timeout,
        returns None.

        Exceptions raised by inferiors are propagated normally, but it is possible for an exception to be delayed
        until the next invocation of this method.
        """
        # First of all, handle pending errors, because removing the item from the queue might unblock reader tasks.
        try:
            exc = self._error_queue.get_nowait()
        except asyncio.QueueEmpty:
            pass
        else:
            assert not isinstance(exc, (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError))
            raise exc
        # Check the read queue only if there are no pending errors.
        loop = asyncio.get_running_loop()
        try:
            timeout = monotonic_deadline - loop.time()
            if timeout > 0:
                tr = await asyncio.wait_for(self._read_queue.get(), timeout)
            else:
                tr = self._read_queue.get_nowait()
        except (asyncio.TimeoutError, asyncio.QueueEmpty):
            # If there are unprocessed transfers, allow the caller to read them even if the instance is closed.
            if self._finalizer is None:
                raise pyuavcan.transport.ResourceClosedError(f"{self} is closed") from None
            return None
        # We do not re-check the error queue at the output because that would mean losing the received transfer.
        # If there are new errors, they will be handled at the next invocation.
        return tr

    @property
    def transfer_id_timeout(self) -> float:
        """
        Assignment of a new transfer-ID timeout is transferred to all inferior sessions,
        so that their settings are always kept consistent.
        When the transfer-ID timeout value is queried, the maximum value from the inferior sessions is returned;
        if there are no inferiors, zero is returned.
        The transfer-ID timeout is not kept by the redundant session itself.

        When a new inferior session is added, its transfer-ID timeout is assigned to match other inferiors.
        When all inferior sessions are removed, the transfer-ID timeout configuration becomes lost.
        Therefore, when the first inferior is added, the redundant session assumes its transfer-ID timeout
        configuration as its own; all inferiors added later will inherit the same setting.
        """
        if self._inferiors:
            return max(x.transfer_id_timeout for x in self.inferiors)
        return 0.0

    @transfer_id_timeout.setter
    def transfer_id_timeout(self, value: float) -> None:
        value = float(value)
        if value <= 0.0:
            raise ValueError(f"Transfer-ID timeout shall be a positive number of seconds, got {value}")
        for s in self.inferiors:
            s.transfer_id_timeout = value

    @property
    def specifier(self) -> pyuavcan.transport.InputSessionSpecifier:
        return self._specifier

    @property
    def payload_metadata(self) -> pyuavcan.transport.PayloadMetadata:
        return self._payload_metadata

    def sample_statistics(self) -> RedundantSessionStatistics:
        """
        - ``transfers``     - the number of successfully received deduplicated transfers (unique transfer count).
        - ``errors``        - the number of receive calls that could not be completed due to an exception.
        - ``payload_bytes`` - the number of payload bytes in successful deduplicated transfers counted in ``transfers``.
        - ``drops``         - the total number of drops summed from all inferiors (i.e., total drop count).
          This value is invalidated when the set of inferiors is changed. The semantics may change later.
        - ``frames``        - the total number of frames summed from all inferiors (i.e., replicated frame count).
          This value is invalidated when the set of inferiors is changed. The semantics may change later.
        """
        inferiors = [s.sample_statistics() for s in self.inferiors]
        return RedundantSessionStatistics(
            transfers=self._stat_transfers,
            frames=sum(s.frames for s in inferiors),
            payload_bytes=self._stat_payload_bytes,
            errors=self._stat_errors,
            drops=sum(s.drops for s in inferiors),
            inferiors=inferiors,
        )

    def close(self) -> None:
        for inf in self._inferiors:
            try:
                inf.close()
            except Exception as ex:
                _logger.exception("%s: Could not close %s: %s", self, inf, ex)
        self._inferiors.clear()
        fin, self._finalizer = self._finalizer, None
        if fin is not None:
            fin()
        self._deduplicator = None

    async def _process_transfer(
        self, session: pyuavcan.transport.InputSession, transfer: pyuavcan.transport.TransferFrom
    ) -> None:
        assert self._deduplicator is not None
        iface_id = id(session)
        if self._deduplicator.should_accept_transfer(
            iface_id=iface_id,
            transfer_id_timeout=self.transfer_id_timeout,
            timestamp=transfer.timestamp,
            source_node_id=transfer.source_node_id,
            transfer_id=transfer.transfer_id,
        ):
            _logger.debug("%s: Accepting %s from %016x", self, transfer, iface_id)
            self._stat_transfers += 1
            self._stat_payload_bytes += sum(map(len, transfer.fragmented_payload))
            await self._read_queue.put(
                RedundantTransferFrom(
                    timestamp=transfer.timestamp,
                    priority=transfer.priority,
                    transfer_id=transfer.transfer_id,
                    fragmented_payload=transfer.fragmented_payload,
                    source_node_id=transfer.source_node_id,
                    inferior_session=session,
                )
            )
        else:
            _logger.debug("%s: Discarding redundant duplicate %s from %016x", self, transfer, iface_id)

    async def _inferior_worker_task(self, session: pyuavcan.transport.InputSession) -> None:
        iface_id = id(session)
        loop = asyncio.get_running_loop()
        try:
            _logger.debug("%s: Task for inferior %016x is starting", self, iface_id)
            while self._deduplicator is not None:
                try:
                    deadline = loop.time() + RedundantInputSession._READ_TIMEOUT
                    tr = await session.receive(deadline)
                    if tr is not None and self._deduplicator is not None:
                        await self._process_transfer(session, tr)
                except (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError):
                    break
                except Exception as ex:
                    # We block until the error is stored in the one-element error queue.
                    # This behavior allows us to avoid spinning broken inferiors that raise errors continuously.
                    _logger.debug("%s: Receive from %016x raised %s", self, iface_id, ex, exc_info=True)
                    self._stat_errors += 1
                    await self._error_queue.put(ex)
        except (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError):
            pass
        except Exception as ex:
            _logger.exception("%s: Task for %016x has encountered an unhandled exception: %s", self, iface_id, ex)
        finally:
            _logger.debug("%s: Task for %016x is stopping", self, iface_id)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

# noinspection PyUnresolvedReferences
r"""
Application layer overview
++++++++++++++++++++++++++

The application module contains the application-layer API.
This module is not imported automatically because it depends on the transpiled DSDL namespace ``uavcan``.
The DSDL namespace can be either transpiled manually or lazily ad-hoc; see :mod:`pyuavcan.dsdl` for related docs.


Node class
++++++++++

The abstract class :class:`pyuavcan.application.Node` models a UAVCAN node ---
it is one of the main entities of the library, along with its factory :meth:`make_node`.
The application uses its Node instance to interact with the network:
create publications/subscriptions, invoke and serve RPC-services.


Constructing a node
^^^^^^^^^^^^^^^^^^^

..  doctest::
    :hide:

    >>> import os
    >>> os.environ["UAVCAN__NODE__ID"]                   = "42"
    >>> os.environ["UAVCAN__PUB__MEASURED_VOLTAGE__ID"]  = "6543"
    >>> os.environ["UAVCAN__SUB__POSITION_SETPOINT__ID"] = "6544"
    >>> os.environ["UAVCAN__SRV__LEAST_SQUARES__ID"]     = "123"
    >>> os.environ["UAVCAN__CLN__LEAST_SQUARES__ID"]     = "123"
    >>> os.environ["UAVCAN__LOOPBACK"]                   = "1"
    >>> import tests
    >>> tests.asyncio_allow_event_loop_access_from_top_level()
    >>> from tests import doctest_await

Create a node using the factory :meth:`make_node` and start it:

>>> import pyuavcan.application
>>> import uavcan.node                                  # Transcompiled DSDL namespace (see pyuavcan.dsdl).
>>> node_info = pyuavcan.application.NodeInfo(          # This is an alias for uavcan.node.GetInfo.Response.
...     software_version=uavcan.node.Version_1(major=1, minor=0),
...     name="org.uavcan.pyuavcan.docs",
... )
>>> node = pyuavcan.application.make_node(node_info)    # Some of the fields in node_info are set automatically.
>>> node.start()

The node instance we just started will periodically publish ``uavcan.node.Heartbeat`` and ``uavcan.node.port.List``,
respond to ``uavcan.node.GetInfo`` and ``uavcan.register.Access``/``uavcan.register.List``,
and do some other standard things -- read the docs for :class:`Node` for details.

Now we can create ports --- that is, instances of
:class:`pyuavcan.presentation.Publisher`,
:class:`pyuavcan.presentation.Subscriber`,
:class:`pyuavcan.presentation.Client`,
:class:`pyuavcan.presentation.Server`
--- to interact with the network.
To create a new port you need to specify its type and name
(the name can be omitted if a fixed port-ID is defined for the data type).


Publishers and subscribers
^^^^^^^^^^^^^^^^^^^^^^^^^^

Create a publisher and publish a message (here and below, ``doctest_await`` substitutes for the ``await`` statement):

>>> import uavcan.si.unit.voltage
>>> pub_voltage = node.make_publisher(uavcan.si.unit.voltage.Scalar_1, "measured_voltage")
>>> pub_voltage.publish_soon(uavcan.si.unit.voltage.Scalar_1(402.15))            # Publish message asynchronously.
>>> doctest_await(pub_voltage.publish(uavcan.si.unit.voltage.Scalar_1(402.15)))  # Or synchronously.
True

Create a subscription and receive a message from it:

..  doctest::
    :hide:

    >>> import uavcan.si.unit.length
    >>> pub = node.presentation.make_publisher(uavcan.si.unit.length.Vector3_1, 6544)
    >>> pub.publish_soon(uavcan.si.unit.length.Vector3_1([42.0, 15.4, -8.7]))

>>> import uavcan.si.unit.length
>>> sub_position = node.make_subscriber(uavcan.si.unit.length.Vector3_1, "position_setpoint")
>>> msg, metadata = doctest_await(sub_position.receive_for(timeout=0.5))
>>> msg.meter[0], msg.meter[1], msg.meter[2]                            # Some payload in the message we received.
(42.0, 15.4, -8.7)
>>> metadata.source_node_id, metadata.priority, metadata.transfer_id    # Metadata for the message.
(42, <Priority.NOMINAL: 4>, 0)


RPC-service clients and servers
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Define an RPC-service of an application-specific type:

>>> from sirius_cyber_corp import PerformLinearLeastSquaresFit_1    # An application-specific DSDL definition.
>>> async def solve_linear_least_squares(                           # Refer to the Demo chapter for the DSDL sources.
...     request: PerformLinearLeastSquaresFit_1.Request,
...     metadata: pyuavcan.presentation.ServiceRequestMetadata,
... ) -> PerformLinearLeastSquaresFit_1.Response:                   # Business logic.
...     import numpy as np
...     x = np.array([p.x for p in request.points])
...     y = np.array([p.y for p in request.points])
...     s, *_ = np.linalg.lstsq(np.vstack([x, np.ones(len(x))]).T, y, rcond=None)
...     return PerformLinearLeastSquaresFit_1.Response(slope=s[0], y_intercept=s[1])
>>> srv_least_squares = node.get_server(PerformLinearLeastSquaresFit_1, "least_squares")
>>> srv_least_squares.serve_in_background(solve_linear_least_squares)  # Run the server in a background task.

Invoke the service we defined above assuming that it is served by node 42:

>>> from sirius_cyber_corp import PointXY_1
>>> cln_least_sq = node.make_client(PerformLinearLeastSquaresFit_1, 42, "least_squares")
>>> req = PerformLinearLeastSquaresFit_1.Request([PointXY_1(10, 1), PointXY_1(20, 2)])
>>> response, metadata = doctest_await(cln_least_sq.call(req))
>>> round(response.slope, 1), round(response.y_intercept, 1)
(0.1, 0.0)

Here is another example showcasing the use of a standard service with a fixed port-ID:

>>> client_node_info = node.make_client(uavcan.node.GetInfo_1, 42)    # Port name is not required.
>>> response, metadata = doctest_await(client_node_info.call(uavcan.node.GetInfo_1.Request()))
>>> response.software_version
uavcan.node.Version.1.0(major=1, minor=0)


Registers and application settings
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

You are probably wondering, how come we just created a node without specifying which transport it should use,
its node-ID, or even the subject-IDs and service-IDs?
Where did these values come from?

They were read from from the *registry* --- a key-value configuration parameter storage [#parameter_server]_
defined in the UAVCAN Specification, chapter *Application layer*, section *Register interface*.
The factory :meth:`make_node` we used above just reads the registers and figures out how to construct
the node from that: which transport to use, the node-ID, the subject-IDs, and so on.
Any UAVCAN application is also expected to keep its own configuration parameters in the registers so that
it can be reconfigured and controlled at runtime via UAVCAN.

The registry of the local node can be accessed via :attr:`Node.registry` which is an instance of class
:class:`pyuavcan.application.register.Registry`:

>>> int(node.registry["uavcan.node.id"])        # Standard registers defined by UAVCAN are named like "uavcan.*"
42
>>> node.id                                     # Yup, indeed, the node-ID is picked up from the register.
42
>>> int(node.registry["uavcan.pub.measured_voltage.id"])    # This is where we got the subject-ID from.
6543
>>> pub_voltage.port_id
6543
>>> int(node.registry["uavcan.sub.position_setpoint.id"])   # And so on.
6544
>>> str(node.registry["uavcan.sub.position_setpoint.type"]) # Port types are automatically exposed via registry, too.
'uavcan.si.unit.length.Vector3.1.0'

Every port created by the application (publisher, subscriber, etc.) is automatically exposed via the register
interface as prescribed by the Specification [#avoid_presentation_layer]_.

New registers (application-specific registers in particular) can be created using
:meth:`pyuavcan.application.register.Registry.setdefault`:

>>> from pyuavcan.application.register import Value, Real64  # Convenience aliases for uavcan.register.Value, etc.
>>> gains = node.registry.setdefault("my_app.controller.pid_gains", Real64([1.3, 0.8, 0.05]))   # Explicit real64 here.
>>> gains.floats
[1.3, 0.8, 0.05]
>>> import numpy as np
>>> node.registry.setdefault("my_app.estimator.state_vector",       # Not stored, but computed at every invocation.
...                          lambda: np.random.random(4)).floats    # Deduced type: real64.
[..., ..., ..., ...]

But the above does not explain where did the example get the register values from.
There are two places:

- **The register file** which contains a simple key-value database table.
  If the file does not exist (like at the first run), it is automatically created.
  If no file location is provided when invoking :meth:`make_node`,
  the registry is stored in memory so that all state is lost when the node is closed.

- **The environment variables.**
  A register like ``m.motor.inductance_dq`` can be assigned via environment variable ``M__MOTOR__INDUCTANCE_DQ``
  (the mapping is documented in the standard RPC-service ``uavcan.register.Access``).
  The value of an environment variable is a space-separated list of values (in case of arrays), or a plain string.
  The environment variables are checked once when the node is constructed, and also whenever a new register is
  created using :meth:`pyuavcan.application.register.Registry.setdefault`.

..  doctest::
    :hide:

    >>> node.close()
    >>> import os
    >>> for k in os.environ:
    ...     if "__" in k:
    ...         del os.environ[k]
    >>> os.environ["UAVCAN__NODE__ID"]                   = "42"
    >>> os.environ["UAVCAN__PUB__MEASURED_VOLTAGE__ID"]  = "6543"
    >>> os.environ["UAVCAN__SUB__OPTIONAL_PORT__ID"]     = "65535"
    >>> os.environ["UAVCAN__UDP__IFACE"]                 = "127.63.0.0"
    >>> os.environ["UAVCAN__SERIAL__IFACE"]              = "socket://127.0.0.1:50905"
    >>> os.environ["UAVCAN__DIAGNOSTIC__SEVERITY"]       = "3.1"
    >>> os.environ["M__MOTOR__INDUCTANCE_DQ"]            = "0.12 0.13"

>>> import os
>>> for k in os.environ:  # Suppose that the following environment variables were passed to our process:
...     if "__" in k:
...         print(k.ljust(40), os.environ[k])
UAVCAN__NODE__ID                         42
UAVCAN__PUB__MEASURED_VOLTAGE__ID        6543
UAVCAN__SUB__OPTIONAL_PORT__ID           65535
UAVCAN__UDP__IFACE                       127.63.0.0
UAVCAN__SERIAL__IFACE                    socket://127.0.0.1:50905
UAVCAN__DIAGNOSTIC__SEVERITY             3.1
M__MOTOR__INDUCTANCE_DQ                  0.12 0.13
>>> node = pyuavcan.application.make_node(node_info, "registers.db")  # The file will be created if doesn't exist.
>>> node.id
42
>>> node.presentation.transport     # Heterogeneously redundant transport: UDP+Serial, as specified in env vars.
RedundantTransport(UDPTransport('127.63.0.42', ...), SerialTransport('socket://127.0.0.1:50905', ...))
>>> pub_voltage = node.make_publisher(uavcan.si.unit.voltage.Scalar_1, "measured_voltage")
>>> pub_voltage.port_id
6543
>>> int(node.registry["uavcan.diagnostic.severity"])                            # This is a standard register.
3
>>> node.registry.setdefault("m.motor.inductance_dq", [1.23, -8.15]).floats     # The value is taken from environment!
[0.12, 0.13]
>>> node.registry.setdefault("m.motor.flux_linkage_dq", [1.23, -8.15]).floats   # No environment variable for this one.
[1.23, -8.15]
>>> node.registry["m.motor.inductance_dq"] = [1.9, 6]                           # Assign new value.
>>> node.registry["m.motor.inductance_dq"].floats
[1.9, 6.0]
>>> node.make_subscriber(uavcan.si.unit.voltage.Scalar_1, "optional_port")      # doctest: +IGNORE_EXCEPTION_DETAIL
Traceback (most recent call last):
...
PortNotConfiguredError: 'uavcan.sub.optional_port.id'
>>> node.close()

..  doctest::
    :hide:

    >>> for k in os.environ:
    ...     if "__" in k:
    ...         del os.environ[k]
    >>> node.close()        # Ensure idempotency.

Per the Specification, a port-ID of 65535 (0xFFFF) represents an unconfigured port,
as illustrated in the above snippet.


Application-layer function implementations
++++++++++++++++++++++++++++++++++++++++++

As mentioned in the description of the Node class, it provides certain bare-minumum standard application-layer
functionality like publishing heartbeats, responding to GetInfo, serving the register API, etc.
More complex capabilities are to be set up by the user as needed; some of them are:

..  autosummary::
    pyuavcan.application.diagnostic.DiagnosticSubscriber
    pyuavcan.application.node_tracker.NodeTracker
    pyuavcan.application.plug_and_play.Allocatee
    pyuavcan.application.plug_and_play.Allocator
    pyuavcan.application.file.FileServer
    pyuavcan.application.file.FileClient


..  [#parameter_server]
    Those familiar with ROS may find similarities with the *ROS Parameter Server*,
    except that each node keeps its own registers locally instead of relying on a remote centralized provider.

..  [#avoid_presentation_layer]
    The application therefore should not attempt to create new ports using the presentation-layer API because that
    would circumvent the introspection services.
"""

from ._node import Node as Node, NodeInfo as NodeInfo, PortNotConfiguredError as PortNotConfiguredError

from ._node_factory import make_node as make_node

from ._transport_factory import make_transport as make_transport

from ._registry_factory import make_registry as make_registry

from . import register as register


class NetworkTimeoutError(TimeoutError):
    """
    API calls below the application layer return None on timeout.
    Some of the application-layer API calls raise this exception instead.
    """



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import typing
import dataclasses
import pyuavcan.transport
from ._base import Deduplicator


class MonotonicDeduplicator(Deduplicator):
    def __init__(self) -> None:
        self._remote_states: typing.List[typing.Optional[_RemoteState]] = []

    def should_accept_transfer(
        self,
        *,
        iface_id: int,
        transfer_id_timeout: float,
        timestamp: pyuavcan.transport.Timestamp,
        source_node_id: typing.Optional[int],
        transfer_id: int,
    ) -> bool:
        del iface_id  # Not used in monotonic deduplicator.
        if source_node_id is None:
            # Anonymous transfers are fully stateless, so always accepted.
            # This may lead to duplications and reordering but this is a design limitation.
            return True

        # If a similar architecture is used on an embedded system, this normally would be a static array.
        if len(self._remote_states) <= source_node_id:
            self._remote_states += [None] * (source_node_id - len(self._remote_states) + 1)
            assert len(self._remote_states) == source_node_id + 1

        if self._remote_states[source_node_id] is None:
            # First transfer from this node, create new state and accept unconditionally.
            self._remote_states[source_node_id] = _RemoteState(last_transfer_id=transfer_id, last_timestamp=timestamp)
            return True

        # We have seen transfers from this node before, so we need to perform actual deduplication.
        state = self._remote_states[source_node_id]
        assert state is not None

        # If we have seen transfers with higher TID values recently, reject this one as duplicate.
        tid_timeout = (timestamp.monotonic - state.last_timestamp.monotonic) > transfer_id_timeout
        if not tid_timeout and transfer_id <= state.last_transfer_id:
            return False

        # Otherwise, this is either a new transfer or a TID timeout condition has occurred.
        state.last_transfer_id = transfer_id
        state.last_timestamp = timestamp
        return True


@dataclasses.dataclass
class _RemoteState:
    last_transfer_id: int
    last_timestamp: pyuavcan.transport.Timestamp



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

import types
import typing
import pkgutil
import importlib


T = typing.TypeVar("T", bound=object)  # https://github.com/python/mypy/issues/5374


def iter_descendants(ty: typing.Type[T]) -> typing.Iterable[typing.Type[T]]:
    # noinspection PyTypeChecker,PyUnresolvedReferences
    """
    Returns a recursively descending iterator over all subclasses of the argument.

    >>> class A: pass
    >>> class B(A): pass
    >>> class C(B): pass
    >>> class D(A): pass
    >>> set(iter_descendants(A)) == {B, C, D}
    True
    >>> list(iter_descendants(D))
    []
    >>> bool in set(iter_descendants(int))
    True

    Practical example -- discovering what transports are available:

    >>> import pyuavcan
    >>> pyuavcan.util.import_submodules(pyuavcan.transport)
    >>> list(sorted(map(lambda t: t.__name__, pyuavcan.util.iter_descendants(pyuavcan.transport.Transport))))
    [...'CANTransport'...'RedundantTransport'...'SerialTransport'...]
    """
    # noinspection PyArgumentList
    for t in ty.__subclasses__():
        yield t
        yield from iter_descendants(t)


def import_submodules(
    root_module: types.ModuleType, error_handler: typing.Optional[typing.Callable[[str, ImportError], None]] = None
) -> None:
    # noinspection PyTypeChecker,PyUnresolvedReferences
    """
    Recursively imports all submodules and subpackages of the specified Python module or package.
    This is mostly intended for automatic import of all available specialized implementations
    of a certain functionality when they are spread out through several submodules which are not
    auto-imported.

    :param root_module: The module to start the recursive descent from.

    :param error_handler: If None (default), any :class:`ImportError` is raised normally,
        thereby terminating the import process after the first import error (e.g., a missing dependency).
        Otherwise, this would be a function that is invoked whenever an import error is encountered
        instead of raising the exception. The arguments are:

        - the name of the parent module whose import could not be completed due to the error;
        - the culprit of type :class:`ImportError`.

    >>> import pyuavcan
    >>> pyuavcan.util.import_submodules(pyuavcan.transport)  # One missing dependency would fail everything.
    >>> pyuavcan.transport.loopback.LoopbackTransport
    <class 'pyuavcan.transport.loopback...LoopbackTransport'>

    >>> import tests.util.import_error  # For demo purposes, this package contains a missing import.
    >>> pyuavcan.util.import_submodules(tests.util.import_error)  # Yup, it fails.
    Traceback (most recent call last):
      ...
    ModuleNotFoundError: No module named 'nonexistent_module_should_raise_import_error'
    >>> pyuavcan.util.import_submodules(tests.util.import_error,  # The handler allows us to ignore ImportError.
    ...                                 lambda parent, ex: print(parent, ex.name))
    tests.util.import_error._subpackage nonexistent_module_should_raise_import_error
    """
    for _, module_name, _ in pkgutil.walk_packages(root_module.__path__, root_module.__name__ + "."):
        try:
            importlib.import_module(module_name)
        except ImportError as ex:
            if error_handler is None:
                raise
            error_handler(module_name, ex)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
This module does not implement a transport, and it is not a part of the abstract transport model.
It contains a collection of software components implementing common logic reusable
with different transport implementations.
It is expected that some transport implementations may be unable to rely on these.

This module is unlikely to be useful for a regular library user (not a developer).
"""

from . import crc as crc
from . import high_overhead_transport as high_overhead_transport

from ._refragment import refragment as refragment



# Copyright (c) 2020 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

"""
This module implements forwarding between the standard subject ``uavcan.diagnostic.Record``
and Python's standard logging facilities (:mod:`logging`).
"""

from __future__ import annotations
import sys
import asyncio
import logging
from typing import Optional
from uavcan.diagnostic import Record_1 as Record
from uavcan.diagnostic import Severity_1 as Severity
import pyuavcan
import pyuavcan.application


__all__ = ["DiagnosticSubscriber", "DiagnosticPublisher", "Record", "Severity"]


_logger = logging.getLogger(__name__)


class DiagnosticSubscriber:
    """
    Subscribes to ``uavcan.diagnostic.Record`` and forwards every received message into Python's :mod:`logging`.
    The logger name is that of the current module.
    The log level mapping is defined by :attr:`SEVERITY_UAVCAN_TO_PYTHON`.

    This class is convenient for various CLI tools and automation scripts where the user will not
    need to implement additional logic to see log messages from the network.
    """

    SEVERITY_UAVCAN_TO_PYTHON = {
        Severity.TRACE: logging.INFO,
        Severity.DEBUG: logging.INFO,
        Severity.INFO: logging.INFO,
        Severity.NOTICE: logging.INFO,
        Severity.WARNING: logging.WARNING,
        Severity.ERROR: logging.ERROR,
        Severity.CRITICAL: logging.CRITICAL,
        Severity.ALERT: logging.CRITICAL,
    }

    def __init__(self, node: pyuavcan.application.Node):
        sub_record = node.make_subscriber(Record)
        node.add_lifetime_hooks(
            lambda: sub_record.receive_in_background(self._on_message),
            sub_record.close,
        )

    async def _on_message(self, msg: Record, meta: pyuavcan.transport.TransferFrom) -> None:
        node_id = meta.source_node_id if meta.source_node_id is not None else "anonymous"
        diag_text = msg.text.tobytes().decode("utf8", errors="replace")
        log_text = (
            f"uavcan.diagnostic.Record: node={node_id} severity={msg.severity.value} "
            + f"ts_sync={msg.timestamp.microsecond * 1e-6:0.6f} ts_local={meta.timestamp}:\n"
            + diag_text
        )
        level = self.SEVERITY_UAVCAN_TO_PYTHON.get(msg.severity.value, logging.CRITICAL)
        _logger.log(level, log_text)


class DiagnosticPublisher(logging.Handler):
    # noinspection PyTypeChecker,PyUnresolvedReferences
    """
    Implementation of :class:`logging.Handler` that forwards all log messages via the standard
    diagnostics subject of UAVCAN.
    Log messages that are too long to fit into a UAVCAN Record object are truncated.
    Log messages emitted by PyUAVCAN itself may be dropped to avoid infinite recursion.

    Here's a usage example. Set up test rigging:

    ..  doctest::
        :hide:

        >>> import tests
        >>> _ = tests.dsdl.compile()
        >>> tests.asyncio_allow_event_loop_access_from_top_level()
        >>> from tests import doctest_await

    >>> from pyuavcan.transport.loopback import LoopbackTransport
    >>> from pyuavcan.application import make_node, NodeInfo, make_registry
    >>> node = make_node(NodeInfo(), transport=LoopbackTransport(1))
    >>> node.start()

    Instantiate publisher and install it with the logging system:

    >>> diagnostic_pub = DiagnosticPublisher(node, level=logging.INFO)
    >>> logging.root.addHandler(diagnostic_pub)
    >>> diagnostic_pub.timestamping_enabled = True  # This is only allowed if the UAVCAN network uses the wall clock.
    >>> diagnostic_pub.timestamping_enabled
    True

    Test it:

    >>> sub = node.make_subscriber(Record)
    >>> logging.info('Test message')
    >>> msg, _ = doctest_await(sub.receive_for(1.0))
    >>> msg.text.tobytes().decode()
    'root: Test message'
    >>> msg.severity.value == Severity.INFO     # The log level is mapped automatically.
    True

    Don't forget to remove it afterwards:

    >>> logging.root.removeHandler(diagnostic_pub)
    >>> node.close()

    The node factory :func:`pyuavcan.application.make_node` actually allows you to do this automatically,
    so that you don't have to hard-code behaviors in the application sources:

    >>> registry = make_registry(None, {"UAVCAN__DIAGNOSTIC__SEVERITY": "2", "UAVCAN__DIAGNOSTIC__TIMESTAMP": "1"})
    >>> node = make_node(NodeInfo(), registry, transport=LoopbackTransport(1))
    >>> node.start()
    >>> sub = node.make_subscriber(Record)
    >>> logging.info('Test message')
    >>> msg, _ = doctest_await(sub.receive_for(1.0))
    >>> msg.text.tobytes().decode()
    'root: Test message'
    >>> msg.severity.value == Severity.INFO
    True
    >>> node.close()
    """

    def __init__(self, node: pyuavcan.application.Node, level: int = logging.WARNING) -> None:
        self._pub = node.make_publisher(Record)
        self._pub.priority = pyuavcan.transport.Priority.OPTIONAL
        self._pub.send_timeout = 10.0

        self._fut: Optional[asyncio.Future[None]] = None
        self._forward_timestamp = False
        self._started = False
        super().__init__(level)

        def start() -> None:
            self._started = True

        def close() -> None:
            self._started = False
            self._pub.close()
            if self._fut is not None:
                try:
                    self._fut.result()
                except asyncio.InvalidStateError:
                    pass  # May be unset https://github.com/UAVCAN/pyuavcan/issues/192

        node.add_lifetime_hooks(start, close)

    @property
    def timestamping_enabled(self) -> bool:
        """
        If True, the publisher will be setting the field ``timestamp`` of the published log messages to
        :attr:`logging.LogRecord.created` (with the appropriate unit conversion).
        If False (default), published messages will not be timestamped at all.
        """
        return self._forward_timestamp

    @timestamping_enabled.setter
    def timestamping_enabled(self, value: bool) -> None:
        self._forward_timestamp = bool(value)

    def emit(self, record: logging.LogRecord) -> None:
        """
        This method intentionally drops all low-severity messages originating from within PyUAVCAN itself
        to prevent infinite recursion through the logging system.
        """
        if not self._started or (record.module.startswith(pyuavcan.__name__) and record.levelno < logging.ERROR):
            return

        # Further, unconditionally drop all messages while publishing is in progress for the same reason.
        # This logic may need to be reviewed later.
        if self._fut is not None and self._fut.done():
            self._fut.result()
            self._fut = None

        dcs_rec = DiagnosticPublisher.log_record_to_diagnostic_message(record, self._forward_timestamp)
        if self._fut is None:
            self._fut = asyncio.ensure_future(self._publish(dcs_rec))
        else:
            # DROPPED
            pass

    async def _publish(self, record: Record) -> None:
        try:
            if not await self._pub.publish(record):
                print(self, "TIMEOUT", record, file=sys.stderr)  # pragma: no cover
        except pyuavcan.transport.TransportError:
            pass
        except Exception as ex:
            print(self, "ERROR", ex.__class__.__name__, ex, file=sys.stderr)  # pragma: no cover

    @staticmethod
    def log_record_to_diagnostic_message(record: logging.LogRecord, use_timestamp: bool) -> Record:
        from uavcan.time import SynchronizedTimestamp_1 as SynchronizedTimestamp

        ts: Optional[SynchronizedTimestamp] = None
        if use_timestamp:
            ts = SynchronizedTimestamp(microsecond=int(record.created * 1e6))

        # The magic severity conversion formula is found by a trivial linear regression:
        #   Fit[data, {1, x}, {{0, 0}, {10, 1}, {20, 2}, {30, 4}, {40, 5}, {50, 6}}]
        sev = min(7, round(-0.14285714285714374 + 0.12571428571428572 * record.levelno))

        text = f"{record.name}: {record.getMessage()}"
        text = text[:255]  # TODO: this is crude; expose array lengths from DSDL.
        return Record(timestamp=ts, severity=Severity(sev), text=text)

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self._pub)



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import typing
import logging
import asyncio
import dataclasses
import pyuavcan.util
import pyuavcan.dsdl
import pyuavcan.transport
from ._base import MessagePort, MessageClass, PortFinalizer, Closable
from ._error import PortClosedError


# Shouldn't be too large as this value defines how quickly the task will detect that the underlying transport is closed.
_RECEIVE_TIMEOUT = 1


_logger = logging.getLogger(__name__)


#: Type of the async received message handler callable.
ReceivedMessageHandler = typing.Callable[[MessageClass, pyuavcan.transport.TransferFrom], typing.Awaitable[None]]


@dataclasses.dataclass
class SubscriberStatistics:
    transport_session: pyuavcan.transport.SessionStatistics  #: Shared per session specifier.
    messages: int  #: Number of received messages, individual per subscriber.
    overruns: int  #: Number of messages lost to queue overruns; individual per subscriber.
    deserialization_failures: int  #: Number of messages lost to deserialization errors; shared per session specifier.


class Subscriber(MessagePort[MessageClass]):
    """
    A task should request its own independent subscriber instance from the presentation layer controller.
    Do not share the same subscriber instance across different tasks. This class implements the RAII pattern.

    Whenever a message is received from a subject, it is deserialized once and the resulting object is
    passed by reference into each subscriber instance. If there is more than one subscriber instance for
    a subject, accidental mutation of the object by one consumer may affect other consumers. To avoid this,
    the application should either avoid mutating received message objects or clone them beforehand.

    This class implements the async iterator protocol yielding received messages.
    Iteration stops shortly after the subscriber is closed.
    It can be used as follows::

        async for message, transfer in subscriber:
            ...  # Handle the message.
        # The loop will be stopped shortly after the subscriber is closed.

    Implementation info: all subscribers sharing the same session specifier also share the same
    underlying implementation object containing the transport session which is reference counted and destroyed
    automatically when the last subscriber with that session specifier is closed;
    the user code cannot access it and generally shouldn't care.
    """

    def __init__(self, impl: SubscriberImpl[MessageClass], queue_capacity: typing.Optional[int]):
        """
        Do not call this directly! Use :meth:`Presentation.make_subscriber`.
        """
        assert not impl.is_closed, "Internal logic error"
        if queue_capacity is None:
            queue_capacity = 0  # This case is defined by the Queue API. Means unlimited.
        else:
            queue_capacity = int(queue_capacity)
            if queue_capacity < 1:
                raise ValueError(f"Invalid queue capacity: {queue_capacity}")

        self._closed = False
        self._impl = impl
        self._maybe_task: typing.Optional[asyncio.Task[None]] = None
        self._rx: _Listener[MessageClass] = _Listener(asyncio.Queue(maxsize=queue_capacity))
        impl.add_listener(self._rx)

    # ----------------------------------------  HANDLER-BASED API  ----------------------------------------

    def receive_in_background(self, handler: ReceivedMessageHandler[MessageClass]) -> None:
        """
        Configures the subscriber to invoke the specified handler whenever a message is received.
        The handler is an async callable or returns an awaitable.

        If the caller attempts to configure multiple handlers by invoking this method repeatedly,
        only the last configured handler will be active (the old ones will be forgotten).
        If the handler throws an exception, it will be suppressed and logged.

        This method internally starts a new task. If the subscriber is closed while the task is running,
        the task will be silently cancelled automatically; the application need not get involved.

        This method of handling messages should not be used with the plain async receive API;
        an attempt to do so may lead to unpredictable message distribution between consumers.
        """

        async def task_function() -> None:
            # This could be an interesting opportunity for optimization: instead of using the queue, just let the
            # implementation class invoke the handler from its own receive task directly. Eliminates extra indirection.
            while not self._closed:
                try:
                    async for message, transfer in self:
                        try:
                            await handler(message, transfer)
                        except Exception as ex:
                            if isinstance(ex, asyncio.CancelledError):
                                raise
                            _logger.exception("%s got an unhandled exception in the message handler: %s", self, ex)
                except (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError) as ex:
                    _logger.debug("%s receive task is stopping because: %r", self, ex)
                    break
                except Exception as ex:
                    _logger.exception("%s receive task failure: %s", self, ex)
                    await asyncio.sleep(1)  # TODO is this an adequate failure management strategy?

        if self._maybe_task is not None:
            self._maybe_task.cancel()

        self._maybe_task = asyncio.get_event_loop().create_task(task_function())

    # ----------------------------------------  DIRECT RECEIVE  ----------------------------------------

    async def receive(
        self, monotonic_deadline: float
    ) -> typing.Optional[typing.Tuple[MessageClass, pyuavcan.transport.TransferFrom]]:
        """
        Blocks until either a valid message is received,
        in which case it is returned along with the transfer which delivered it;
        or until the specified deadline is reached, in which case None is returned.
        The deadline value is compared against :meth:`asyncio.AbstractEventLoop.time`.

        The method will never return None unless the deadline has been exceeded or the session is closed;
        in order words, a spurious premature return cannot occur.

        If the deadline is not in the future, the method will non-blockingly check if there is any data;
        if there is, it will be returned, otherwise None will be returned immediately.
        It is guaranteed that no context switch will occur in this case, as if the method was not async.

        If an infinite deadline is desired, consider using :meth:`__aiter__`/:meth:`__anext__`.
        """
        loop = asyncio.get_running_loop()
        return await self.receive_for(timeout=monotonic_deadline - loop.time())

    async def receive_for(
        self, timeout: float
    ) -> typing.Optional[typing.Tuple[MessageClass, pyuavcan.transport.TransferFrom]]:
        """
        This is like :meth:`receive` but with a relative timeout instead of an absolute deadline.
        """
        self._raise_if_closed_or_failed()
        try:
            if timeout > 0:
                message, transfer = await asyncio.wait_for(self._rx.queue.get(), timeout)
            else:
                message, transfer = self._rx.queue.get_nowait()
        except asyncio.QueueEmpty:
            return None
        except asyncio.TimeoutError:
            return None
        else:
            assert isinstance(message, self._impl.dtype), "Internal protocol violation"
            assert isinstance(transfer, pyuavcan.transport.TransferFrom), "Internal protocol violation"
            return message, transfer

    # ----------------------------------------  ITERATOR API  ----------------------------------------

    def __aiter__(self) -> Subscriber[MessageClass]:
        """
        Iterator API support. Returns self unchanged.
        """
        return self

    async def __anext__(self) -> typing.Tuple[MessageClass, pyuavcan.transport.TransferFrom]:
        """
        This is like :meth:`receive` with an infinite timeout, so it cannot return None.
        """
        try:
            while not self._closed:
                out = await self.receive_for(_RECEIVE_TIMEOUT)
                if out is not None:
                    return out
        except pyuavcan.transport.ResourceClosedError:
            pass
        raise StopAsyncIteration

    # ----------------------------------------  AUXILIARY  ----------------------------------------

    @property
    def dtype(self) -> typing.Type[MessageClass]:
        return self._impl.dtype

    @property
    def transport_session(self) -> pyuavcan.transport.InputSession:
        return self._impl.transport_session

    def sample_statistics(self) -> SubscriberStatistics:
        """
        Returns the statistical counters of this subscriber, including the statistical metrics of the underlying
        transport session, which is shared across all subscribers with the same session specifier.
        """
        return SubscriberStatistics(
            transport_session=self.transport_session.sample_statistics(),
            messages=self._rx.push_count,
            deserialization_failures=self._impl.deserialization_failure_count,
            overruns=self._rx.overrun_count,
        )

    def close(self) -> None:
        if not self._closed:
            self._closed = True
            self._impl.remove_listener(self._rx)
            if self._maybe_task is not None:  # The task may be holding the lock.
                try:
                    self._maybe_task.cancel()  # We don't wait for it to exit because it's pointless.
                except Exception as ex:
                    _logger.exception("%s task could not be cancelled: %s", self, ex)
                self._maybe_task = None

    def _raise_if_closed_or_failed(self) -> None:
        if self._closed:
            raise PortClosedError(repr(self))

        if self._rx.exception is not None:
            self._closed = True
            raise self._rx.exception from RuntimeError("The subscriber has failed and been closed")

    def __del__(self) -> None:
        try:
            closed = self._closed
        except AttributeError:
            closed = True  # Incomplete construction.
        if not closed:
            # https://docs.python.org/3/reference/datamodel.html#object.__del__
            # DO NOT invoke logging from the finalizer because it may resurrect the object!
            # Once it is resurrected, we may run into resource management issue if __del__() is invoked again.
            # Whether it is invoked the second time is an implementation detail.
            self._closed = True
            self._impl.remove_listener(self._rx)


@dataclasses.dataclass
class _Listener(typing.Generic[MessageClass]):
    """
    The queue-induced extra level of indirection adds processing overhead and latency. In the future we may need to
    consider an optimization where the subscriber would automatically detect whether the underlying implementation
    is shared among many subscribers or not. If not, it should bypass the queue and read from the transport directly
    instead. This would avoid the unnecessary overheads and at the same time would be transparent for the user.
    """

    queue: asyncio.Queue[typing.Tuple[MessageClass, pyuavcan.transport.TransferFrom]]
    push_count: int = 0
    overrun_count: int = 0
    exception: typing.Optional[Exception] = None

    def push(self, message: MessageClass, transfer: pyuavcan.transport.TransferFrom) -> None:
        try:
            self.queue.put_nowait((message, transfer))
            self.push_count += 1
        except asyncio.QueueFull:
            self.overrun_count += 1

    def __repr__(self) -> str:
        """
        Overriding repr() is necessary to avoid the contents of the queue from being printed.
        The queue contains DSDL objects, which may be large and the output of their repr() may be very expensive
        to compute, especially if the queue is long.
        """
        return pyuavcan.util.repr_attributes_noexcept(
            self,
            queue_length=self.queue.qsize(),
            push_count=self.push_count,
            overrun_count=self.overrun_count,
            exception=self.exception,
        )


class SubscriberImpl(Closable, typing.Generic[MessageClass]):
    """
    This class implements the actual reception and deserialization logic. It is not visible to the user and is not
    part of the API. There is at most one instance per session specifier. It may be shared across multiple users
    with the help of the proxy class. When the last proxy is closed or garbage collected, the implementation will
    also be closed and removed.
    """

    def __init__(
        self,
        dtype: typing.Type[MessageClass],
        transport_session: pyuavcan.transport.InputSession,
        finalizer: PortFinalizer,
    ):
        self.dtype = dtype
        self.transport_session = transport_session
        self.deserialization_failure_count = 0
        self._maybe_finalizer: typing.Optional[PortFinalizer] = finalizer
        self._task = asyncio.get_event_loop().create_task(self._task_function())
        self._listeners: typing.List[_Listener[MessageClass]] = []

    @property
    def is_closed(self) -> bool:
        return self._maybe_finalizer is None

    async def _task_function(self) -> None:
        exception: typing.Optional[Exception] = None
        loop = asyncio.get_running_loop()
        try:  # pylint: disable=too-many-nested-blocks
            while not self.is_closed:
                transfer = await self.transport_session.receive(loop.time() + _RECEIVE_TIMEOUT)
                if transfer is not None:
                    message = pyuavcan.dsdl.deserialize(self.dtype, transfer.fragmented_payload)
                    if message is not None:
                        for rx in self._listeners:
                            rx.push(message, transfer)
                    else:
                        self.deserialization_failure_count += 1
        except (asyncio.CancelledError, pyuavcan.transport.ResourceClosedError) as ex:
            _logger.debug("Cancelling the subscriber task of %s because: %r", self, ex)
        except Exception as ex:
            exception = ex
            # Do not use f-string because it can throw, unlike the built-in formatting facility of the logger
            _logger.exception("Fatal error in the subscriber task of %s: %s", self, ex)
        finally:
            self._finalize(exception)

    def _finalize(self, exception: typing.Optional[Exception] = None) -> None:
        exception = exception if exception is not None else PortClosedError(repr(self))
        try:
            if self._maybe_finalizer is not None:
                self._maybe_finalizer([self.transport_session])
                self._maybe_finalizer = None
        except Exception as ex:
            _logger.exception("Failed to finalize %s: %s", self, ex)
        for rx in self._listeners:
            rx.exception = exception

    def close(self) -> None:
        try:
            self._task.cancel()  # Force the task to be stopped ASAP without waiting for timeout
        except Exception as ex:
            _logger.debug("Explicit close: could not cancel the task %r: %s", self._task, ex, exc_info=True)
        self._finalize()

    def add_listener(self, rx: _Listener[MessageClass]) -> None:
        assert not self.is_closed, "Internal logic error: cannot add listener to a closed subscriber implementation"
        self._listeners.append(rx)

    def remove_listener(self, rx: _Listener[MessageClass]) -> None:
        try:
            self._listeners.remove(rx)
        except ValueError:
            _logger.exception("%r does not have listener %r", self, rx)
        if len(self._listeners) == 0:
            self.close()

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes_noexcept(
            self,
            dtype=str(pyuavcan.dsdl.get_model(self.dtype)),
            transport_session=self.transport_session,
            deserialization_failure_count=self.deserialization_failure_count,
            listeners=self._listeners,
            closed=self.is_closed,
        )



# Copyright (c) 2019 UAVCAN Consortium
# This software is distributed under the terms of the MIT License.
# Author: Pavel Kirienko <pavel@uavcan.org>

from __future__ import annotations
import abc
import typing
import pyuavcan.util
import pyuavcan.dsdl
import pyuavcan.transport


DEFAULT_PRIORITY = pyuavcan.transport.Priority.NOMINAL
"""
This value is not mandated by Specification, it is an implementation detail.
"""

DEFAULT_SERVICE_REQUEST_TIMEOUT = 1.0
"""
This value is recommended by Specification.
"""

PortFinalizer = typing.Callable[[typing.Sequence[pyuavcan.transport.Session]], None]


TypeClass = typing.TypeVar("TypeClass", bound=pyuavcan.dsdl.CompositeObject)
MessageClass = typing.TypeVar("MessageClass", bound=pyuavcan.dsdl.CompositeObject)
ServiceClass = typing.TypeVar("ServiceClass", bound=pyuavcan.dsdl.ServiceObject)


class OutgoingTransferIDCounter:
    """
    A member of the output transfer-ID map. Essentially this is just a boxed integer.
    The value is monotonically increasing starting from zero;
    transport-specific modulus is computed by the underlying transport(s).
    """

    def __init__(self) -> None:
        """
        Initializes the counter to zero.
        """
        self._value: int = 0

    def get_then_increment(self) -> int:
        """
        Samples the counter with post-increment; i.e., like ``i++``.
        """
        out = self._value
        self._value += 1
        return out

    def override(self, value: int) -> None:
        """
        Assigns a new value. Raises a :class:`ValueError` if the value is not a non-negative integer.
        """
        value = int(value)
        if value >= 0:
            self._value = value
        else:
            raise ValueError(f"Not a valid transfer-ID value: {value}")

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(self, self._value)


class Closable(abc.ABC):
    """
    Base class for closable session resources.
    """

    @abc.abstractmethod
    def close(self) -> None:
        """
        Invalidates the object and closes the underlying resources if necessary.

        If the closed object had a blocked task waiting for data, the task will raise a
        :class:`pyuavcan.presentation.PortClosedError` shortly after close;
        or, if the task was started by the closed instance itself, it will be silently cancelled.
        At the moment the library provides no guarantees regarding how quickly the exception will be raised
        or the task cancelled; it is only guaranteed that it will happen automatically eventually, the
        application need not be involved in that.
        """
        raise NotImplementedError


class Port(Closable, typing.Generic[TypeClass]):
    """
    The base class for any presentation layer session such as publisher, subscriber, client, or server.
    The term "port" came to be from <https://forum.uavcan.org/t/a-generic-term-for-either-subject-or-service/182>.
    """

    @property
    @abc.abstractmethod
    def dtype(self) -> typing.Type[TypeClass]:
        """
        The generated Python class modeling the corresponding DSDL data type.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def port_id(self) -> int:
        """
        The immutable subject-/service-ID of the underlying transport session instance.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __repr__(self) -> str:
        raise NotImplementedError


# noinspection DuplicatedCode
class MessagePort(Port[MessageClass]):
    """
    The base class for publishers and subscribers.
    """

    @property
    @abc.abstractmethod
    def transport_session(self) -> pyuavcan.transport.Session:
        """
        The underlying transport session instance. Input for subscribers, output for publishers.
        One instance per session specifier.
        """
        raise NotImplementedError

    @property
    def port_id(self) -> int:
        ds = self.transport_session.specifier.data_specifier
        assert isinstance(ds, pyuavcan.transport.MessageDataSpecifier)
        return ds.subject_id

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self, dtype=str(pyuavcan.dsdl.get_model(self.dtype)), transport_session=self.transport_session
        )


# noinspection DuplicatedCode
class ServicePort(Port[ServiceClass]):
    @property
    @abc.abstractmethod
    def input_transport_session(self) -> pyuavcan.transport.InputSession:
        """
        The underlying transport session instance used for the input transfers
        (requests for servers, responses for clients). One instance per session specifier.
        """
        raise NotImplementedError

    @property
    def port_id(self) -> int:
        ds = self.input_transport_session.specifier.data_specifier
        assert isinstance(ds, pyuavcan.transport.ServiceDataSpecifier)
        return ds.service_id

    def __repr__(self) -> str:
        return pyuavcan.util.repr_attributes(
            self, dtype=str(pyuavcan.dsdl.get_model(self.dtype)), input_transport_session=self.input_transport_session
        )
